/*---------------------------------------------------------------------------
 *
 *      Title:          Farsi Text Editor - Amiga specific module
 *
 *      File:           amiga.c
 *
 *      Date:           May 1995
 *
 *      Description:
 *          This module implements the Amiga machine specific functions.
 *
 *      Rev     By              Description                            Date
 *      1.0     R.J.Harrison    Implement                              2-5-95
 *
 *      Copyright  1995 Zaretto Ltd.
 *---------------------------------------------------------------------------*/


#include <exec/types.h>
#include <libraries/dos.h>
#include <workbench/workbench.h>
#include <workbench/startup.h>
#include <proto/icon.h>
#ifdef LATTICE
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#endif

#include <stdio.h>
#include <unix/getargs.h>

/*
 * The following are both fudges...
 */
int _stack_size = 43212;
int bss_size = 10000;

#include "zpl.h"

#define VERSION "1.11"
#define VERSION_DATE "12-June-95"
#define VERSION_TYPE ""

/*
 * _main.c         Copyright (C) 1985  Lattice, Inc.       
 */
#include <fcntl.h>
#include <ios1.h>
#include <time.h>
#include <workbench/startup.h>
#include <libraries/dosextens.h>
#include <proto/dos.h>
#include <proto/exec.h>
#include <exec/execbase.h>

extern TY_int do_scan_check;
extern void tell_em ();

TY_int from_wb = TRUE;

ARG arg[] =
{
    {
	'm', ARG_TBOOLEAN, &do_scan_check, "Scan memory for overwrites every 5 seconds"
    }
};

static char *version_string = "$VER:FARSI_EDITOR_" VERSION "-" VERSION_TYPE "(" VERSION_DATE ")";

char *get_version (void)
{
    return ("V" VERSION " " VERSION_TYPE);
}

void usage (void)
{
    fprintf (stderr, "Farsi: Farsi Text Editor %s, 1995 Zaretto Ltd.\n", get_version ());
}

static APTR old_pr_win_ptr = NULL;
static struct Process *proc = NULL;

/*
 * Normal termination.
 */
void terminate (TY_int return_code)
{
    PR_save_database (PR_get_prefs ());
    PR_reset_changes ();
    PR_finish ();

    if (proc->pr_WindowPtr != old_pr_win_ptr)
    {
	proc = (struct Process *) FindTask (NULL);
	proc->pr_WindowPtr = old_pr_win_ptr;
    }

    close_system ();
    MY_ov_check_memory ();
    exit (0);
}

/*
 * Perform the minimum closedown to cleanup and exit...
 */
void guru_exit (void)
{
    if (proc->pr_WindowPtr != old_pr_win_ptr)
    {
	proc = (struct Process *) FindTask (NULL);
	proc->pr_WindowPtr = old_pr_win_ptr;
    }
    close_system ();
    Delay (50);			/*
				 * Waiting a while 
				 */
}

TY_int _CXBRK (void)
{
    static TY_int already_terminating = FALSE;
    printf ("**Interrupt**\n");
    if (!already_terminating)
    {
	already_terminating = TRUE;
	terminate (0);
    }
    else
	printf ("CTRL-C pressed whilst already being processed\n");
    return (0);
}				/*
				 * Disable Lattice CTRL/C handling 
				 */

TY_int CXBRK (void)
{
    _CXBRK ();
    return (0);
}

static void sigint_handler (TY_int code)
{
    _CXBRK ();
    exit (0);
}

/*
 * Opens and allocations we must clean up 
 */
struct Library *IconBase = NULL;
#define conwin stdout
static LONG olddir = -1;

BOOL parse_tool_types (struct WBArg *wbarg)
{
    struct DiskObject *dobj;
    char **toolarray;
    char *s;
    BOOL success = FALSE;

    if ((*wbarg->wa_Name) && (dobj = GetDiskObject (wbarg->wa_Name)))
    {
	toolarray = (char **) dobj->do_ToolTypes;

	if (s = (char *) FindToolType (toolarray, "PAGE"))
	{
	}
#ifdef AAA
	if (s = (char *) FindToolType (toolarray, "FLAGS"))
	{
	    fprintf (conwin, "    Found tooltype FLAGS with value %s\n", s);
	    if (MatchToolValue (s, "BOLD"))
		fprintf (conwin, "      BOLD flag requested\n");
	    if (MatchToolValue (s, "ITALICS"))
		fprintf (conwin, "      ITALICS flag requested\n");
	}
#endif
	/*
	 * Free the diskobject we got 
	 */
	FreeDiskObject (dobj);
	success = TRUE;
    }
#ifdef AAA
    else if (!(*wbarg->wa_Name))
	fprintf (conwin, "  Must be a disk or drawer icon\n");
    else
	fprintf (conwin, "  Can't find DiskObject (icon) for this WBArg\n");
#endif
    return (success);
}


void WB_do_args (char **argv)
{
    struct WBStartup *WBenchMsg;
    struct WBArg *wbarg;
    SHORT i;

    /*
     * Open icon.library 
     */
    if (!(IconBase = OpenLibrary ("icon.library", 33)))
    {
	printf ("Can't open icon.library\n", RETURN_FAIL);
	return;
    }
    WBenchMsg = (struct WBStartup *) argv;

    /*
     * Note wbarg++ at end of FOR statement steps through wbargs.
     * * First arg is our executable (tool).  Any additional args
     * * are projects/icons passed to us via either extend select
     * * or default tool method.
     */
    for (i = 0, wbarg = WBenchMsg->sm_ArgList;
	 i < WBenchMsg->sm_NumArgs;
	 i++, wbarg++)
    {
	/*
	 * if there's a directory lock for this wbarg, CD there 
	 */
	olddir = -1;
	if ((wbarg->wa_Lock) && (*wbarg->wa_Name))
	    olddir = CurrentDir (wbarg->wa_Lock);

	parse_tool_types (wbarg);

#ifdef __O_OQW_EOWQE
	if ((i > 0) && (*wbarg->wa_Name))
	    fprintf (conwin, "In Main. We could open the %s file here\n",
		     wbarg->wa_Name);
#endif
	if (olddir != -1)
	    CurrentDir (olddir);	/*
					 * CD back where we were 
					 */
    }
    if (IconBase)
	CloseLibrary (IconBase);
}



TY_int init_machine (TY_int argc, char **argv)
{
    signal (SIGINT, sigint_handler);

    proc = (struct Process *) FindTask (NULL);
    old_pr_win_ptr = proc->pr_WindowPtr;

    /*
     * Process any arguments, either from the command line, or from the
     * workbench tooltypes.
     */
    if (argc)
    {
	from_wb = FALSE;
	argc = getargs (argc, argv, arg, sizeof (arg) / sizeof (ARG));
	argc--;
	argv++;
    }
    else
	WB_do_args (argv);

    return (0);
}

/*
 * Replace the standard SAS/C Code for _main, to provide the required extra
 * functionality that we need.
 */
extern struct ExecBase *SysBase;

#define MAXARG 32
#define QUOTE       '"'
#define MAXWINDOW   40
#define ESCAPE '*'
#define ESC_CODE '\027'
#define NL '\n'

#define isspace(c)      ((c == ' ')||(c == '\t') || (c == '\n'))

extern TY_int __fmode;

extern char __stdiowin[];
extern char __stdiov37[];

extern struct WBStartup *_WBenchMsg;

static TY_int argc;		/*
				 * arg count 
				 */
static char **targv, *argv[MAXARG + 1];		/*
						 * arg pointers 
						 */

void __stdargs __main (line)
char *line;
{
    char **pargv;
    char *argbuf;
    TY_int ret;

    TY_int x;
    struct Process *process;
    struct FileHandle *handle;
    struct UFB ufbs[3];


/***
 *     Build argument pointer list
 ***/
    while (argc < MAXARG)
    {
	while (isspace (*line))
	    line++;
	if (*line == '\0')
	    break;
	pargv = &argv[argc++];
	if (*line == QUOTE)
	{
	    argbuf = *pargv = ++line;	/*
					 * ptr inside quoted string 
					 */
	    while (*line != QUOTE && *line != 0)
	    {
		if (*line == ESCAPE)
		{
		    line++;
		    switch (*line)
		    {
		    case '\0':
			*argbuf = 0;
			goto linedone;
		    case 'E':
			*argbuf++ = ESC_CODE;
			break;
		    case 'N':
			*argbuf++ = NL;
			break;
		    default:
			*argbuf++ = *line;
		    }
		    line++;
		}
		else
		{
		    *argbuf++ = *line++;
		}
	    }
	    line++;
	    *argbuf++ = '\0';	/*
				 * terminate arg 
				 */
	}
	else
	    /*
	     * non-quoted arg 
	     */
	{
	    *pargv = line;
	    while ((*line != '\0') && (!isspace (*line)))
		line++;
	    if (*line == '\0')
		break;
	    else
		*line++ = '\0';	/*
				 * terminate arg 
				 */
	}
    }				/*
				 * while 
				 */

  linedone:

    targv = (argc == 0) ? (char **) _WBenchMsg : (char **) &argv[0];


/***
 *     Open standard files
 ***/

    ufbs[0].ufbnxt = &ufbs[1];
    ufbs[1].ufbnxt = &ufbs[2];
    ufbs[2].ufbnxt = NULL;
    __ufbs = &ufbs[0];
    ufbs[0].ufbfn = NULL;
    ufbs[1].ufbfn = NULL;
    ufbs[2].ufbfn = NULL;
    ufbs[0].ufbflg = UFB_RA | O_RAW | UFB_NC;
    ufbs[1].ufbflg = UFB_WA | O_RAW | UFB_NC;
    ufbs[2].ufbflg = UFB_WA | O_RAW | UFB_NC | UFB_CLO;

    if (argc == 0)
    {
	/*
	 * running under workbench      
	 */
	ufbs[0].ufbfh = Open (get_debug_filename (), MODE_NEWFILE);

	if (ufbs[0].ufbfh == NULL)
	    ufbs[0].ufbfh = Open ("NIL:", MODE_NEWFILE);

	ufbs[0].ufbflg |= UFB_CLO;
	handle = (struct FileHandle *) (ufbs[0].ufbfh << 2);
	process = (struct Process *) FindTask (0);
	process->pr_ConsoleTask = (APTR) handle->fh_Type;
	if ((ufbs[1].ufbfh = Open ("*", MODE_OLDFILE)) == NULL)
	    ufbs[1].ufbfh = Open ("NIL:", MODE_OLDFILE);
	ufbs[2].ufbfh = ufbs[1].ufbfh;
    }
    else
    {				/*
				 * running under CLI            
				 */
	ufbs[0].ufbfh = Input ();
	ufbs[1].ufbfh = Output ();
	if ((ufbs[2].ufbfh = Open ("*", MODE_OLDFILE)) == NULL)
	    ufbs[2].ufbfh = Open ("NIL:", MODE_OLDFILE);
    }


    __nufbs += 3;


    x = (__fmode) ? 0 : _IOXLAT;
    stdin->_file = 0;
    stdin->_flag = _IOREAD | x;
    stdout->_file = 1;
    stdout->_flag = _IOWRT | _IOLBF | x;
    stderr->_file = 2;
    stderr->_flag = _IOWRT | _IOLBF | x;


/***
 *     Call user's main program
 ***/

    ret = main (argc, (char **) targv);		/*
						 * call main function 
						 */

    exit (ret);
}

static void 
AlterAppKeyMap (struct MY_KeyMap *appkeymap)
{
appkeymap->km_LoKeyMap[0].alt       = 0x90;
appkeymap->km_LoKeyMap[0].shift_alt = 0x90;

appkeymap->km_LoKeyMap[0x2b].plain     = '#';

/*
 * Disable the shift-3 key
 */
appkeymap->km_LoKeyMap[3].shift = 0;
appkeymap->km_LoKeyMap[3].shift_alt = 0;

//appkeymap->km_LoKeyMapTypes[3] = KCF_NOP;
//appkeymap->km_LoKeyMapTypes[3] = KCF_NOP;

}

void init_keyboard(void)
{
set_event_keymap (CreateAppKeyMap ( AlterAppKeyMap ));
}

