/*---------------------------------------------------------------------------
 *
 *      Title:          Farsi Text Editor - Farsi Text Engine
 *
 *      File:           amiga.c
 *
 *      Date:           May 1995
 *
 *      Description:
 *           This module contains the intelligence for handling the language
 *      together with handling command line input given key inputs
 *
 *      Rev     By              Description                            Date
 *      1.0     R.J.Harrison    Implement                              2-5-95
 *      1.1     R.J.Harrison    Fix end characters, add numeric ins   13-5-95
 *      1.2     R.J.Harrison    Fix # and ,                           17-5-95
 *      1.3     R.J.Harrison    Add numeric deletion                  18-5-95
 *      1.4     R.J.Harrison    Page support added                    20-5-95 
 *      1.5     R.J.Harrison    Fix cursor movement after load        12-6-95 
 *
 *      Copyright  1995 Zaretto Ltd.
 *---------------------------------------------------------------------------*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "zpl.h"

TRAN_TABLE     tran_table[CHARSET_SIZE];
TRAN_TABLE     tails[CHARSET_SIZE];
/*

 * The special cases are to allow re use of the character set,
 * but also allowing different attributes (e.g. a tail)
 */
char           special_case[CHARSET_SIZE];
char           tail_char[CHARSET_SIZE];

char           is_farsi[CHARSET_SIZE];

/*
 * Import the translation and exception tables
 */

EXCEPTION      exceptions[] =
{
    {44, ET_SA},
#ifdef FARSI_NUMERICS
    {49, ET_SA},
    {50, ET_SA},
    {51, ET_SA},
    {52, ET_SA},
    {53, ET_SA},
    {54, ET_SA},
    {55, ET_SA},
    {56, ET_SA},
    {57, ET_SA},
    {48, ET_SA},
#endif
    {67, ET_SA},
    {72, ET_SA},
    {90, ET_SA},
    {92, ET_SA},
    {98, ET_SA},
    {99, ET_SA},
    {104, ET_SA},
    {110, ET_SA},
    {118, ET_SA},
    {0, ET_END}
};

#define GET_KEY(tt_key,ar) if (tt_key) ar = tt_key

/*
 * Return the standalone character index in tran_table 
 * given any character.
 */
int
locate_base_index (int ch)
{
    static int     back_trans[CHARSET_SIZE], init = 0;

    if (!init)
    {
	int            i;

	init = 1;

	memset (back_trans, 0, sizeof (back_trans));

	for (i = 0; i < CHARSET_SIZE; i++)
	{
	    GET_KEY (i, back_trans[tran_table[i].stand]);
	    GET_KEY (i, back_trans[tran_table[i].beg]);
	    GET_KEY (i, back_trans[tran_table[i].mid]);
	    GET_KEY (i, back_trans[tran_table[i].end]);
	}
    }
    return back_trans[ch];
}

int
locate_base (int ch)
{
    return tran_table[locate_base_index (ch)].stand;
}

void 
init_tran_table (int ch, int stand, int beg, int mid, int end)
{
    tran_table[ch].stand = stand;
    tran_table[ch].beg = beg;
    tran_table[ch].mid = mid;
    tran_table[ch].end = end;
}

void 
add_tail (int ch, int stand, int beg, int mid, int end)
{
    tails[tran_table[ch].stand].stand = stand;	/* Mark this as having a tail */
    tails[tran_table[ch].beg].beg = beg;	/* Mark this as having a tail */
    tails[tran_table[ch].mid].mid = mid;	/* Mark this as having a tail */
    tails[tran_table[ch].end].end = end;	/* Mark this as having a tail */
    tail_char[stand] = 1;
    tail_char[beg] = 1;
    tail_char[mid] = 1;
    tail_char[end] = 1;
}

void
farsi_init (void)
{
    memset (tran_table, 0, sizeof (tran_table));
    memset (tails, 0, sizeof (tails));
    memset (special_case, 0, sizeof (special_case));

    init_tran_table (0x90, 96,96,96,96);

    init_tran_table ('`', 126, 39, 39, 126);
    add_tail ('`', 58, 0, 0, 58);

#ifdef FARSI_NUMERICS
    init_tran_table ('1', 49, 49, 49, 49);
    init_tran_table ('2', 50, 50, 50, 50);
    init_tran_table ('3', 51, 51, 51, 51);
    init_tran_table ('4', 52, 52, 52, 52);
    init_tran_table ('5', 53, 53, 53, 53);
    init_tran_table ('6', 54, 54, 54, 54);
    init_tran_table ('7', 55, 55, 55, 55);
    init_tran_table ('8', 56, 56, 56, 56);
    init_tran_table ('9', 57, 57, 57, 57);
    init_tran_table ('0', 48, 48, 48, 48);
#endif

    init_tran_table ('q', 81, 113, 113, 81);
    add_tail ('q', 124, 0, 0, 124);

    init_tran_table ('w', 87, 119, 119, 87);
    add_tail ('w', 124, 0, 0, 124);

    init_tran_table ('e', 69, 101, 101, 69);
    add_tail ('e', 58, 0, 0, 58);

    init_tran_table ('r', 82, 114, 114, 82);
    init_tran_table ('t', 116, 116, 116, 116);
    add_tail ('t', 58, 0, 0, 58);

    init_tran_table ('y', 89, 121, 66, 86);
    init_tran_table ('u', 85, 117, 78, 77);
    init_tran_table ('i', 73, 105, 105, 84);
    init_tran_table ('o', 79, 111, 111, 79);
    init_tran_table ('p', 80, 112, 112, 80);
    init_tran_table ('[', 91, 123, 123, 91);
    init_tran_table (']', 93, 125, 125, 93);
    init_tran_table ('a', 65, 97, 97, 65);
    add_tail ('a', 124, 0, 0, 124);

    init_tran_table ('s', 83, 115, 115, 83);
    add_tail ('s', 124, 0, 0, 124);

    init_tran_table ('d', 68, 100, 100, 88);
    init_tran_table ('f', 70, 102, 102, 70);
    add_tail ('f', 58, 0, 0, 58);

    init_tran_table ('g', 71, 103, 103, 71);
    init_tran_table ('h', 104, 104, 72, 72);
    init_tran_table ('j', 74, 106, 106, 74);
    add_tail ('j', 58, 0, 0, 58);

    init_tran_table ('k', 75, 107, 107, 75);
    init_tran_table ('l', 76, 108, 108, 76);

    init_tran_table (';', 59, 59, 59, 59);
    add_tail (';', 58, 0, 0, 58);

    init_tran_table ('#', 35, 35, 35, 35);
    add_tail ('#', 58, 0, 0, 58);

    init_tran_table ('\\', 92, 92, 92, 92);
    init_tran_table ('z', 122, 122, 122, 122);
    init_tran_table ('x', 120, 120, 120, 120);
    init_tran_table ('c', 99, 99, 99, 99);
    init_tran_table ('v', 118, 118, 118, 118);
    init_tran_table ('b', 98, 98, 98, 98);
    init_tran_table ('n', 110, 110, 110, 110);
    init_tran_table ('m', 109, 109, 109, 109);
    add_tail ('m', 58, 0, 0, 58);

    init_tran_table (',', 44, 44, 44, 44);
    init_tran_table ('_', 95, 95, 95, 95);
    init_tran_table (':', 94, 94, 94, 94);
    init_tran_table ('Z', 90, 90, 90, 90);
    init_tran_table ('C', 67, 67, 67, 90);

    memset (is_farsi, TRUE, sizeof (is_farsi));

    is_farsi[0] = 0;

    is_farsi[32] = 0;
    is_farsi[33] = 0;
    is_farsi[34] = 0;
    is_farsi[36] = 0;
    is_farsi[37] = 0;
    is_farsi[38] = 0;
    is_farsi[40] = 0;
    is_farsi[41] = 0;
    is_farsi[42] = 0;
    is_farsi[43] = 0;
    is_farsi[45] = 0;
    is_farsi[46] = 0;
    is_farsi[47] = 0;

    is_farsi[48] = 0;
    is_farsi[49] = 0;
    is_farsi[50] = 0;
    is_farsi[51] = 0;
    is_farsi[52] = 0;
    is_farsi[53] = 0;
    is_farsi[54] = 0;
    is_farsi[55] = 0;
    is_farsi[56] = 0;
    is_farsi[57] = 0;

    is_farsi[60] = 0;
    is_farsi[61] = 0;
    is_farsi[62] = 0;
    is_farsi[63] = 0;
    is_farsi[94] = 0;
    is_farsi[96] = 0;

    locate_base (0);
}

int
locate_exception_index (int ch)
{
    int            i;

    for (i = 0; exceptions[i].code && (exceptions[i].code >= i); i++)
    {
	if (exceptions[i].code == ch)
	    return i;
    }
    return -1;
}

/*
 * Get the correct character code given the base character and the
 * character type required.
 */
int
get_ch (int base, CHAR_TYPE ct)
{
    switch (ct)
    {
    case CT_STAND:
	{
	    return tran_table[base].stand;
	    break;
	}

    case CT_BEG:
	{
	    return tran_table[base].beg;
	    break;
	}

    case CT_MID:
	{
	    return tran_table[base].mid;
	    break;
	}

    case CT_END:
	{
	    return tran_table[base].end;
	    break;
	}
    }
    return tran_table[base].stand;
}

/*
 * Get the correct character code given the base character and the
 * character type required.
 */
int
get_tail (int base, CHAR_TYPE ct)
{
    switch (ct)
    {
    case CT_STAND:
	{
	    return tails[base].stand;
	    break;
	}

    case CT_BEG:
	{
	    return tails[base].beg;
	    break;
	}

    case CT_MID:
	{
	    return tails[base].mid;
	    break;
	}

    case CT_END:
	{
	    return tails[base].end;
	    break;
	}
    }
    return tails[base].stand;
}

/*
 * Figure out whether the current character type
 */
CHAR_TYPE
get_char_type (char *line, int pos)
{
    int            ei;
    CHAR_TYPE      ct = CT_STAND;

    if (pos > 0)
    {
	ei = locate_exception_index (line[pos - 1]);

	if (ei >= 0)
	{
	    if (exceptions[ei].et == ET_SA)
	    {
		if (is_farsi[line[pos + 1]])
		    return CT_BEG;
		else
		    return CT_STAND;
	    }
	}
    }

    if (pos >= (MAXLEN - 1))
    {
	return CT_STAND;
    }

/*
 * When followed by a non-blank : Beginning
 */
    if (!pos || (!is_farsi[line[pos - 1]]))
    {
	if (is_farsi[line[pos + 1]])
	    ct = CT_BEG;
    }
    else
    {
	/*
	 * When preceeded and followed by a non-blank : Middle
	 */
	if (pos
	    && (is_farsi[line[pos - 1]])
	    && (is_farsi[line[pos + 1]]))
	    ct = CT_MID;
    }

/*
 * (always) when followed by a space : end
 */
    if (!is_farsi[line[pos + 1]])
    {
	if (pos && (is_farsi[line[pos - 1]]))
	    return CT_END;
    }

/*
 * When surrounded by spaces : Standalone (default)
 */

    return ct;
}

/*
 * Modify the a specific character position 
 */
void
modify_one_char (char *line, int pos, CHAR_TYPE * ct_line)
{
    int            base_key;
    CHAR_TYPE      ct;

    if (line[pos] == SPACE || !line[pos])
	return;

    /*
     * firstly change the previous character.
     */
    base_key = locate_base_index (line[pos]);
    ct = get_char_type (line, pos);
    ct_line[pos] = ct;

    if (base_key)
	line[pos] = get_ch (base_key, ct);
}

/*
 * Fix numbers as entered to be right to left...
 */
void 
fix_number (char *line, int pos)
{
    int            i;
    /*
     * Only start to modify the numerics to be left to right when
     * a second digit is entered
     */
    if (pos && isdigit (line[pos - 1]))
    {
	char           new_num = line[pos];

	for (i = pos; isdigit (line[i - 1]) && (i > 0); i--)
	    line[i] = line[i - 1];

	line[i] = new_num;
    }
}


/*
 * Modify the character position (and possible before and after positions),
 * depending on the character at pos.
 * the current character will change depending on what is preceeding it.
 * 
 * The preceeding and next characters will change depending on what the
 * current character is
 */

void
modify_char (char *line, int pos, CHAR_TYPE * ct_line)
{
    if (pos)
	modify_one_char (line, pos - 1, ct_line);

    modify_one_char (line, pos, ct_line);

    modify_one_char (line, pos + 1, ct_line);
}

/*
 * Setup the elements of the text engine to reflect the current line.
 */
void 
setup_line (char *line, CHAR_TYPE * ct_line, char *cursor_offset)
{
    int            i, offset_count = 0;

    memset (ct_line, 0, sizeof (*ct_line));

    for (i = 0; i < MAXLEN; i++)
    {
	modify_one_char (line, i, ct_line);

if (cursor_offset)
{
if (get_tail (line[i], ct_line[i]))
	{
	    offset_count++;
	}
        cursor_offset[i] = offset_count;
}
   }
}

/*
 * Convert a line from the internal format to an output ready format.
 */
void 
convert_from_input (char *line)
{
    char          *p = line, *op;
    char           top[INPUT_LINE_SIZE];
    int            i, o_pos;

    memset (top, 0, sizeof (top));

    o_pos = 0;

    for (i = MAXLEN - 1; line[i] && i >= 0; i--)
    {
	top[o_pos++] = line[i];
    }

    p = op = top;
    while (*p)
    {
	if (!tail_char[*p])
	    *op++ = *p;
	p++;
    }
    while (op < p)
	*op++ = 0;

    strcpy (line, top);
    decr (line);		/* Remove trailing spaces */
}

/*
 * Convert a line from the internal format to an output ready format.
 */
void 
convert_to_output (ZPL_WINDOW * fw, char *top, char *line)
{
    char          *cursor_offset = CURSOR_OFFSET_PTR (fw->sub_window);
    CHAR_TYPE      ct, *ct_line = CHARTYPE_PTR (fw->sub_window);
    int            i, o_pos, i_pos, offset_count;

    memset (top, ' ', MAXLEN);

    i_pos = 0;
    o_pos = MAXLEN - 1;
    offset_count = 0;

    for (i = MAXLEN - 1; (i >= 0) && (o_pos >= 0); i--)
    {
	char           ch;

	ch = line[i_pos++];

	if (!ch)
	    top[o_pos--] = ' ';
	else
	{
	    top[o_pos--] = ch + special_case[ch];
	}

	if (ct = get_tail (ch, ct_line[i_pos - 1]))
	{
	    if (o_pos >= 0)
               top[o_pos--] = ct;
	    offset_count++;
	}
	if (cursor_offset)
	    cursor_offset[i_pos - 1] = offset_count;
    }
    cursor_offset[i_pos] = offset_count;

    top[MAXLEN] = 0;
}

void 
edit_cursor_move (ZPL_WINDOW * fw)
{
    EDIT_WINDOW   *ew = (EDIT_WINDOW *) fw->sub_window;
    TEXT_PAGE     *page = TEXT_PAGE_PTR (fw->sub_window);
    char          *cursor_offset = CURSOR_OFFSET_PTR (fw->sub_window);

    FW_draw_cursor (fw, ((fw->input_x + fw->t_width) -
			 fw->input_pos - cursor_offset[fw->input_pos]) - 2,
		    fw->input_y);

    sprintf (ew->status, "Page %4d      | (%2d,%2d)      ", page->page_number,
	     fw->input_y, fw->input_pos + cursor_offset[fw->input_pos] + 1);
}

char          *
set_status (ZPL_WINDOW * fw, char *text, int param)
{
    EDIT_WINDOW   *ew = (EDIT_WINDOW *) fw->sub_window;

    edit_cursor_move (fw);

    sprintf (ew->status, text, param);

    return 0;
}

/*
 * Decide whether the current line is 'full' (depending on the
 * current character).
 */
int line_full(ZPL_WINDOW *fw, char new_ch)
{
    CHAR_TYPE      *ct_line = CHARTYPE_PTR (fw->sub_window);

if (real_position (fw, fw->input_len) >= fw->t_width)
return 1;

    if (get_tail (fw->input_line[fw->input_len-1], ct_line[fw->input_len-1])
        && real_position (fw, fw->input_len)+1 >= fw->t_width)
        return 1;

return 0;

}
/*
 * Line output routine
 * - This will output right to left, all text is stored in the same format in
 *   the internal buffers, however it is output in the expected direction
 *   which could be left to right or right to left
 */
void
line_output (ZPL_WINDOW * fw, int y, char *line)
{
    char           top[INPUT_LINE_SIZE];
    TEXT_PAGE     *page = TEXT_PAGE_PTR (fw->sub_window);

    convert_to_output (fw, top, line);

    memset (page->text[y - 1], 0, 50);
    strmxcpy (page->text[y - 1], line, MAXLEN);

    FW_move (fw, 38, y);
    FW_puts (fw, top);
}

void 
display_page (ZPL_WINDOW * fw)
{
    int            i;
    char           top[INPUT_LINE_SIZE];
    TEXT_PAGE     *page = TEXT_PAGE_PTR (fw->sub_window);
    CHAR_TYPE      *ct_line = CHARTYPE_PTR (fw->sub_window);

    for (i = 0; i < PAGELEN; i++)
    {
        setup_line (page->text[i], ct_line, 0);
	convert_to_output (fw, top, page->text[i]);

	FW_move (fw, 38, i + 1);
	FW_puts (fw, top);
    }
    setup_line (page->text[fw->input_y-1], ct_line, CURSOR_OFFSET_PTR (fw->sub_window));
    edit_cursor_move (fw);
}

/*
 * Input routines (key handling)
 */

int 
real_position (ZPL_WINDOW * fw, int pos)
{
    char          *cursor_offset = CURSOR_OFFSET_PTR (fw->sub_window);

    return pos + cursor_offset[pos];
}

void 
load_line (ZPL_WINDOW * fw)
{
    TEXT_PAGE     *page = TEXT_PAGE_PTR (fw->sub_window);
    CHAR_TYPE      *ct_line = CHARTYPE_PTR (fw->sub_window);
    char          *cursor_offset = CURSOR_OFFSET_PTR (fw->sub_window);
    char           top[INPUT_LINE_SIZE];

    memset (fw->input_line, 0, sizeof (fw->input_line));

    strmxcpy (fw->input_line, page->text[fw->input_y - 1], MAXLEN);
    setup_line (fw->input_line, ct_line, cursor_offset);

    convert_to_output (fw, top, fw->input_line);

    fw->input_len = strlen (fw->input_line);

    if (real_position (fw, fw->input_pos) > real_position(fw,fw->input_len)) 
	fw->input_pos = fw->input_len;

    edit_cursor_move (fw);
}

static         TY_int
process_cursor_key (ZPL_WINDOW * fw, TY_int key_pressed)
{

    switch (key_pressed)
    {
    case CURSOR_UP:
	if (fw->input_y > 1)
	{
	    fw->input_y--;
	    load_line (fw);
	}
	break;

    case CURSOR_DOWN:
	if (fw->input_y < PAGELEN)
	{
	    fw->input_y++;
	    load_line (fw);
	}
	break;

    case CURSOR_RIGHT:
	if (fw->input_pos > 0)
	    fw->input_pos--;
        else if (fw->input_y > 1)
        {
            process_cursor_key (fw,CURSOR_UP);
	    fw->input_pos = fw->input_len;
        }
	break;

    case CURSOR_LEFT:
	if (real_position(fw,fw->input_pos) < real_position(fw,fw->input_len) )
	    fw->input_pos++;
        else if (fw->input_y < PAGELEN)
        {
	    fw->input_pos = 0;
            process_cursor_key (fw,CURSOR_DOWN);
        }
	break;

    default:
	return (FALSE);
    }
    return (TRUE);
}

int 
word_wrap (ZPL_WINDOW * fw, int space_required)
{
    char           cur_ch[200], *ptr = &fw->input_line[fw->input_pos],
                  *p1 = ptr-1;

    if (fw->input_y >= PAGELEN)
	return 0;

    *cur_ch = 0;

    p1--;
    while (p1 > fw->input_line && *p1 && !isspace (*p1))
	p1--;

    if (p1 > fw->input_line)
    {
	*p1++ = 0;
	strcpy (cur_ch, p1);
	while (p1 <= ptr)
	    *p1++ = 0;
    }

    line_output (fw, fw->input_y, fw->input_line);

    fw->input_pos = 0;
    process_cursor_key (fw, CURSOR_DOWN);

    if (*cur_ch)
    {
    /*
     * Insert the wrapped word at the start of the next line, 
     * together with an extra space for the new character.
     * If it doesn't fit, then tough, characters will be lost...
     */
        str_insert (fw->input_line, strlen(cur_ch));
	memcpy (fw->input_line, cur_ch, strlen(cur_ch));
        fw->input_pos = strlen (cur_ch);
        fw->input_len = strlen (fw->input_line);
        setup_line (fw->input_line, CHARTYPE_PTR (fw->sub_window), 0);
    }

    return 1;
}

void
window_input (ZPL_WINDOW * fw, TY_int c)
{
#ifdef AMIGA
    static TY_int  cursor_move = FALSE;
#endif
    TY_int         i, ipos;
    TEXT_PAGE     *page = TEXT_PAGE_PTR (fw->sub_window);

    ipos = fw->input_pos + fw->input_x;
    ipos = min (ipos, fw->t_width - 1);

    fw->lines_printed = 0;

    page->modified = TRUE;
#ifdef AMIGA
    if (c == CSI)
	cursor_move = TRUE;
    else
    {
	if (cursor_move)
	{
	    process_cursor_key (fw, c);
	    edit_cursor_move (fw);

	    cursor_move = FALSE;
	}
	else
#endif
	{
#ifdef WINDOWS
	    if (process_cursor_key (fw, c))
	    {
		edit_cursor_move (fw);
		return;
	    }
#endif

	    FW_colour (fw, PR_get_colour (FOREGROUND_COLOUR));
	    switch (c)
	    {
	    case DEL:
		{
		    if (fw->input_pos != fw->input_len)
		    {
			page->modified = TRUE;
			fw->input_len--;
			for (i = fw->input_pos; i < fw->input_len; i++)
			{
			    fw->input_line[i] = fw->input_line[i + 1];
			}
			fw->input_line[fw->input_len] = 0;
			fw->input_line[fw->input_len + 1] = 0;
			ipos = min (fw->input_pos + fw->input_x, fw->t_width);
                        setup_line (fw->input_line, CHARTYPE_PTR (fw->sub_window), 0);
			modify_char (fw->input_line, fw->input_pos, CHARTYPE_PTR (fw->sub_window));
			modify_char (fw->input_line, fw->input_len, CHARTYPE_PTR (fw->sub_window));
			line_output (fw, fw->input_y, fw->input_line);
			c = 0;
			edit_cursor_move (fw);
			break;
		    }
/*
 * Otherwise treat del as BS by 'falling through to BS'
 */
		}

	    case BS:
		{
		    if (--fw->input_pos < 0)
		    {
 			fw->input_pos = 0;
                        process_cursor_key (fw,CURSOR_RIGHT);
			edit_cursor_move (fw);
			break;
		    }

		    if (--fw->input_len < 0)
		    {
			fw->input_len = 0;
			fw->input_line[0] = 0;
		    }

/*
 * Numeric deletion.
 */
		    if (fw->input_pos && (isdigit (fw->input_line[fw->input_pos]) && isdigit (fw->input_line[fw->input_pos - 1])))
		    {
			int            i = fw->input_pos;
			while ((i >= 0) && isdigit (fw->input_line[--i])) ;
			while (++i <= fw->input_pos)
			{
			    fw->input_line[i] = fw->input_line[i + 1];
			}
		    }

		    for (i = fw->input_pos; i < fw->input_len; i++)
		    {
			fw->input_line[i] = fw->input_line[i + 1];
		    }
		    fw->input_line[fw->input_len] = 0;
		    ipos = min (fw->input_pos + fw->input_x, fw->t_width);
                    setup_line (fw->input_line, CHARTYPE_PTR (fw->sub_window), 0);
		    modify_char (fw->input_line, fw->input_pos, CHARTYPE_PTR (fw->sub_window));
		    line_output (fw, fw->input_y, fw->input_line);
		    c = 0;

		    edit_cursor_move (fw);
		    page->modified = TRUE;
		    break;
		}

	    case EOL:
		fw->input_pos = fw->input_len;
		edit_cursor_move (fw);
		break;

	    case SOL:
		fw->input_pos = 0;
		edit_cursor_move (fw);
		break;

	    case ESC:
//		close_window (fw);
		break;

	    case ERASE_LINE:
		{
		    if (strlen (fw->input_line))
		    {
			page->modified = TRUE;
			fw->input_pos = fw->input_len = 0;
			memset (fw->input_line, 0, sizeof (fw->input_line));
                        setup_line (fw->input_line, CHARTYPE_PTR (fw->sub_window), 0);
			line_output (fw, fw->input_y, fw->input_line);
                        set_status(fw,"Line cleared",0);
		    }
		    break;
		}

	    case RETURN:
		{
		    if (fw->input_y < PAGELEN)
		    {
			fw->input_y++;
			fw->input_pos = 0;
			load_line (fw);
		    }
		    break;
		}
	    default:
		{
		    if (c >= ' ')
		    {
			page->modified = TRUE;
			FW_colour (fw, PR_get_colour (FOREGROUND_COLOUR));
			edit_cursor_move (fw);

			if (tran_table[c].stand < 0)
			    return;	/* Not allowed key */

			if (tran_table[c].stand)
			    c = tran_table[c].stand;	/* Always start with the standalone.. */

			if (real_position (fw, fw->input_pos)+1 >= fw->t_width)
			{
			    if (!word_wrap (fw, 0))
				break;
			}

			if (fw->input_pos != fw->input_len)
			{
			    if ( line_full(fw,c))
			    {
				set_status (fw,"Line full", 0);
				break;
			    }
			    for (i = fw->input_len; i > fw->input_pos; i--)
			    {
				fw->input_line[i] = fw->input_line[i - 1];
			    }
                            fw->input_line[fw->input_pos] = c;
                            setup_line (fw->input_line, CHARTYPE_PTR (fw->sub_window), 0);
			}
			else 
			    fw->input_line[fw->input_pos] = c;


			if (fw->input_pos && (isdigit (c) && isdigit (fw->input_line[fw->input_pos - 1])))
			    fix_number (fw->input_line, fw->input_pos);

			modify_char (fw->input_line, fw->input_pos, CHARTYPE_PTR (fw->sub_window));
			line_output (fw, fw->input_y, fw->input_line);

			fw->input_pos++;
			fw->input_len++;

#ifdef AL
			if (real_position (fw, fw->input_pos) >= fw->t_width)
			{
			    word_wrap (fw, 0);

			    fw->input_pos--;
			    fw->input_len--;
			}
			else
			{
			    if ((real_position (fw, fw->input_len) >= fw->t_width)
				|| (fw->input_len >= fw->t_width))
			    {
				fw->input_pos--;
				fw->input_len--;
			    }
			}
#endif
			edit_cursor_move (fw);

		    }
		}
	    }
	}
#ifdef AMIGA
    }
#endif
}
