/*---------------------------------------------------------------------------
 *
 *      Title:          Farsi Text Editor - Window manager
 *
 *      File:           wm.c
 *
 *      Date:           May 1995
 *
 *      Description:
 *          This module implements the methods for the windows.
 *
 *      Rev     By              Description                            Date
 *      1.0     R.J.Harrison    Implement                              2-5-95
 *
 *      Copyright  1995 Zaretto Ltd.
 *---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef AMIGA
#include <exec/types.h>
#include <exec/nodes.h>
#include <exec/lists.h>
#include <intuition/intuition.h>
#include <intuition/intuitionbase.h>
#include <graphics/text.h>
#include <proto/exec.h>
#include <proto/graphics.h>
#include <libraries/gadtools.h>
#include <proto/intuition.h>
#endif

#include "zpl.h"

#ifdef AMIGA
#define FW_GET_ID(fw) fw
#endif

#ifdef WINDOWS
#define FW_GET_ID(fw) fw
#endif

#define GADGET_IDCMP_FLAGS_ALL		(GADGETDOWN | GADGETUP | MOUSEMOVE | INTUITICKS | VANILLAKEY)

#ifdef AMIGA
#define MW_idcmp 	GADGETUP | GADGETDOWN | REQSET | ACTIVEWINDOW | \
			RAWKEY | MENUPICK | MOUSEBUTTONS | CLOSEWINDOW | ARROWIDCMP | SCROLLERIDCMP | NEWSIZE

#define IW_idcmp 	GADGETUP | GADGETDOWN | MOUSEBUTTONS | CLOSEWINDOW | ARROWIDCMP | SCROLLERIDCMP | NEWSIZE | MENUPICK
#else
#define MW_idcmp  0
#define IW_idcmp  0
#endif

static TY_int ignore_window_prefs;

TY_int process_edit_window (EDIT_WINDOW * pw, ACTION action, TY_int param, TY_int id)
{
    static struct NewWindow pref_window =
{
    1, 1, 640, 246,		/*
				 * start(x,y), Width, height
				 */
    2, 1,			/*
				 * detail, block pens
				 */

 /*
  * IDCMP flags
  */
    MW_idcmp | REFRESHWINDOW,

 /*
  * Window Flags
  */
    WFLG_BACKDROP| WFLG_BORDERLESS | SMART_REFRESH |  ACTIVATE,
    NULL,			/*
				 * Pointer to first gadget
				 */
    NULL,			/*
				 * Pointer to checkmark
				 */
    "RESWIN",			/*
				 * title
				 */
    NULL,			/*
				 * screen pointer
				 */
    NULL,			/*
				 * bitmap pointer
				 */
    ~0, ~0, 10, 10,		/*
				 * window not sized
				 */
    WBENCHSCREEN		/*
				 * type of screen
				 */
};
    char *pref_name = "Main.Window";
    static char *window_name = "Farsi Text Editor";

    PREFERENCE_ENTRY *pe;

    switch (action)
    {
    case WM_ADD_GADGETS:
	{
#ifdef WINDOWS
	    RE_resource *re;
	    if (re = RE_create_resources (FW_GET_ID (pw->hdr.win), "edit"))
	    {
		pw->hdr.win->resources = re;
		MoveWindow (pw->hdr.win->w, FW_get_left_edge(pw->hdr.win), FW_get_top_edge(pw->hdr.win),
		            re->max_x, re->max_y + 40, TRUE);
	    }
	    else
	    {
		error_log ("Unable to locate 'edit' resource in database");
		return (FALSE);
	    }
#endif
            add_console_to_window (pw->hdr.win);

	    if (pw->hdr.win->resources)
		RE_add_resources (pw->hdr.win->resources, pw->hdr.win->w);
	    break;
	}

    case WM_WINDOW_CLOSE:
	{
	    if (pw->ptr1)
                free (pw->ptr1);
	    if (pw->ptr2)
                free (pw->ptr2);
	    if (pw->ptr)
                free (pw->ptr);
	    break;
	}
    case WM_WINDOW_OPEN:
	{
	    /* No menus...init_edit_menu (pw->hdr.win); */ 
	    pw->hdr.win->type |= FW_TYPE_GADTOOLS;
	    UI_edit_init (pw->hdr.win->resources);
	    FW_set_title (pw->hdr.win, window_name);
            RE_set_item (pw->hdr.win->resources, "status", pw->status, sizeof (pw->status));
	    pw->ptr2 = calloc(sizeof(CHAR_TYPE), INPUT_LINE_SIZE);
	    pw->ptr1 = calloc(sizeof(char), INPUT_LINE_SIZE);
	    pw->ptr = calloc(sizeof(TEXT_PAGE), 1);
            draw_bevel_box (pw->hdr.win, 296, 18, 635, 252);
	    break;
	}
    case WM_WINDOW_FINISH:
	{
	}

    case WM_GET_WINDOW_TYPE:
	return (FW_TYPE_EDIT);

    case WM_GET_WINDOW_PREFS:
	{
	    TY_int width = 1, height = 1;
#ifdef AMIGA
	    RE_resource *re;
	    if (re = RE_create_resources (FW_GET_ID (pw->hdr.win), "edit"))
	    {
		pw->hdr.win->resources = re;
		width = re->max_x;
		height = re->max_y;
	    }
	    else
	    {
		error_log ("Unable to locate 'edit' resource in database");
		return (FALSE);

	    }
#else
	    width = 400, height = 300;
#endif
	    pe = PR_get_item (PR_get_prefs (), pref_name);

	    if (!pe)
		PR_add_item (PR_get_prefs (), pref_name, &pref_window, sizeof (pref_window));
	    pe = PR_get_item (PR_get_prefs (), pref_name);

	    if (pe && !ignore_window_prefs)
	    {
/*
 * The title is only a pointer, so we need to reset it ....
 */
		struct NewWindow *wp = (struct NewWindow *) pe->data;
		wp->Title = window_name;
		if (!using_v36 ())
		    wp->Flags |= SMART_REFRESH | NOCAREREFRESH;
		wp->Height = height;
		wp->Width = width;
		//wp->IDCMPFlags |= VANILLAKEY;
		if (using_v36 ())
		    wp->IDCMPFlags |= RAWKEY;
		//wp->IDCMPFlags |= GADGET_IDCMP_FLAGS_ALL;
		return (TY_int) pe->data;
	    }
	    else
	    {
		pref_window.Title = window_name;
		if (!using_v36 ())
		    pref_window.Flags |= SMART_REFRESH | NOCAREREFRESH;
		//pref_window.Flags |= VANILLAKEY;
		return (TY_int) & pref_window;
	    }
	    break;
	}
    case WM_SET_WINDOW_PREFS:
{
	    struct NewWindow *wp = (struct NewWindow *) param;
	    PR_add_item (PR_get_prefs (), pref_name, wp, sizeof (*wp));
	    break;
	}
    case WM_MAKE_WINDOW_PREFS:
	{
	    struct NewWindow *wp;
	    pe = PR_get_item (PR_get_prefs (), pref_name);
	    wp = (struct NewWindow *) pe->data;
	    pref_window = *wp;
	    pref_window.LeftEdge = FW_get_left_edge (pw->hdr.win);
	    pref_window.TopEdge = FW_get_top_edge (pw->hdr.win);
	    pref_window.Width = FW_get_width (pw->hdr.win);
	    pref_window.Height = FW_get_height (pw->hdr.win) - FW_get_border_top (pw->hdr.win);
	    PR_add_item (PR_get_prefs (), pref_name, &pref_window, sizeof (pref_window));
	    break;
	}
    case WM_CONTENT_CHANGED:
	{
	    break;
	}
    case WM_MENU_GROUPS: break; /* Activate menu items */
    }
    return (0);
}

TY_int do_entity (void *_p1, ACTION action, TY_int param, TY_int id)
{
    GENERAL_WINDOW *gw = _p1;
    return (*gw->hdr.process) (gw, action, param, id);
}

GENERAL_WINDOW *make_edit_window (void)
{
    EDIT_WINDOW *gw;

    if (gw = (EDIT_WINDOW *) malloc (sizeof (*gw)))
    {
	extern TY_int process_edit_window ();
	memset (gw, NULL, sizeof (*gw));
	gw->hdr.type = EDIT_WINDOW_ID | LAYOUTABLE;
	gw->hdr.process = process_edit_window;
        gw->hdr.class = "Editor";
	return ((GENERAL_WINDOW *) gw);
    }
    return (NULL);
}

#ifdef WINDOWS
GENERAL_WINDOW *make_MDI_window (void)
{
    INFO_WINDOW *gw;

    if (gw = (INFO_WINDOW *) malloc (sizeof (*gw)))
    {
	extern TY_int process_info_window ();
	memset (gw, NULL, sizeof (*gw));
	gw->hdr.type = INFO_WINDOW_ID;
	gw->hdr.process = process_info_window;
	return ((GENERAL_WINDOW *) gw);
    }
    return (NULL);
}
#endif

