/*---------------------------------------------------------------------------
 *
 *      Title:          NICOLA Comms - Device interface (Amiga)
 *
 *      File:           CM_drv.c
 *
 *      Date:           1994
 *
 *      Description:
 *
 *      Rev: 1.0
 *      1.0     R.J.Harrison    Implement                            1994
 *
 *      Copyright  1994 R.J.Harrison
 *---------------------------------------------------------------------------*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include <exec/types.h>
#include <proto/exec.h>
#include <devices/serial.h>
#include <libraries/dos.h>

#define MODULE "CM_Drv"
#include "OS_def.h"		/*
				 * Os specific definitions 
				 */

#include "MY_alloc.h"		/*
				 * Redefine allocation routines
				 */

#include "comms.h"

void CM_output (COMMS_CONTEXT * context, char *string)
{
    TY_int r;

    context->SerialWriteIO->IOSer.io_Command = CMD_WRITE;
    context->SerialWriteIO->IOSer.io_Length = strlen (string);
    context->SerialWriteIO->IOSer.io_Data = (APTR) string;

    r = DoIO ((struct IORequest *)context->SerialWriteIO);

    if (CM_debug > 1)
	fprintf (CM_debug_win, "OP:%s", string);
}

TY_int CM_input_waiting (COMMS_CONTEXT * context)
{
    TY_int r;
/*
 * Has the outstanding request completed ?
 */
    if (context->input_requested && CheckIO ((struct IORequest *)context->SerialIO))
    {
	WaitIO ((struct IORequest *)context->SerialIO);
	context->input_requested = FALSE;
	return (1);
    }
    context->SerialIO->IOSer.io_Command = SDCMD_QUERY;
    r = DoIO ((struct IORequest *)context->SerialIO);
    return ((TY_int) context->SerialIO->IOSer.io_Actual);
}

void CM_request_input (COMMS_CONTEXT * context)
{
/*
 * Send off a read request.
 */
    if (!context->input_requested)
    {
	context->SerialIO->IOSer.io_Command = CMD_READ;
	context->SerialIO->IOSer.io_Length = 1;
	context->SerialIO->IOSer.io_Data = (APTR) context->input_buffer_1;
	context->SerialIO->IOSer.io_Actual = 0;
	memset (context->input_buffer_1, 0, sizeof (context->input_buffer_1));
	SendIO ((struct IORequest *)context->SerialIO);
	context->input_requested = TRUE;
    }
}

/*
 * This is a low level read from device routine. The when handling code
 * (see above), is used to action any completed when items. Before calling this,
 * the currently read in line will be stored in the when circular buffer
 *
 * The major problem with this routine is mainting asynchronisity. To do this
 * we have to add an activation signal, together with the routine to call to
 * the Window definition. This is (obviously) a function of the comms open
 * routine. This routine has to support slightly 'odd' calling conventions.
 *
 * There is a sub-class for the communications window which contains the
 * comms context. The open routine will only allow one CIX window to be
 * open at any one time. Closing the CIX window will be permitted,
 * however it will cause a 'clean' comms shutdown.
 *
 * - NOTE - All input strings must reside within static memory, as they are
 *          read in asynchronously.
 */

TY_int CM_input (COMMS_CONTEXT * context, char *string)
{
    TY_int r;
    char *buf;
    TY_int lfe;
/*
 * Use the 'high speed operation' technique as described in the RKM.
 * There are three stages to this.
 * 1. Find out how many characters are available to be read (SDCMD_QUERY)
 * 2. Read this many characters via a non-async io (which will return)
 * 3. Post async request for 1 characters, so that when something is ready,
 *    the device will return it (waking us up hopefully).
 */


    if (context->input_requested)
    {
	if (CheckIO ((struct IORequest *)context->SerialIO))
	    WaitIO ((struct IORequest *)context->SerialIO);
	else
	    return (FALSE);

	if (context->SerialIO->IOSer.io_Actual)
	{
	    if (context->logfile)
		lfe = fwrite (context->SerialIO->IOSer.io_Data, context->SerialIO->IOSer.io_Actual, 1, context->logfile);
#ifdef USE_CB
	    CB_add_sized_item (&context->cb, context->SerialIO->IOSer.io_Data, context->SerialIO->IOSer.io_Actual);
#else
	    CM_check_whens (context, context->SerialIO->IOSer.io_Data, context->SerialIO->IOSer.io_Actual);
#endif
	    if (context->fw && !context->do_not_echo)
		FW_write (context->fw, context->SerialIO->IOSer.io_Data, context->SerialIO->IOSer.io_Actual);
	    context->input_requested = FALSE;
	}
    }

    context->SerialIO->IOSer.io_Command = SDCMD_QUERY;
    r = DoIO ((struct IORequest *)context->SerialIO);

/*
 * Check for CD
 */
    context->carrier_detect = !(context->SerialIO->io_Status & (1 << 5));


    if (CM_debug > 1)
	printf ("CM_input: CD %d\n", context->carrier_detect);

/*
 * If there is anything to read, then read it
 */
    if (context->SerialIO->IOSer.io_Actual)
    {
	context->SerialIO->IOSer.io_Command = CMD_READ;
	context->SerialIO->IOSer.io_Length = min (SERIAL_BUFFER_SIZE, context->SerialIO->IOSer.io_Actual);

	if (string)
	    context->SerialIO->IOSer.io_Data = (APTR) string;
	else
	{
	    context->SerialIO->IOSer.io_Data = (APTR) context->input_buffer;
	    memset (context->input_buffer, 0, sizeof (context->input_buffer));
	}

	context->input_requested = FALSE;
	r = DoIO ((struct IORequest *)context->SerialIO);
	if (r)
	    printf ("CM_iw: %d\n", r);

	context->input_size = context->SerialIO->IOSer.io_Actual;
	buf = context->SerialIO->IOSer.io_Data;
	context->buffer_1 = TRUE;

	if (context->logfile)
	    lfe = fwrite (buf, context->input_size, 1, context->logfile);

#ifdef USE_CB
	CB_add_sized_item (&context->cb, context->SerialIO->IOSer.io_Data, context->input_size);
	CM_check_whens (context);
#else
	CM_check_whens (context, buf, context->input_size);
#endif
	if (CM_debug > 1)
	{
	    fwrite (buf, context->input_size, 1, CM_debug_win);
	    fflush (CM_debug_win);
	}
	if (context->fw && !context->do_not_echo)
	    FW_write (context->fw, buf, context->input_size);

	CM_request_input (context);
    }
    else
	CM_request_input (context);
    return (context->input_size);
}

void CM_check_CD (COMMS_CONTEXT * context)
{
    TY_int r;

    if (context->input_requested)
    {
	if (!CheckIO ((struct IORequest *)context->SerialIO))
	{
	    AbortIO ((struct IORequest *)context->SerialIO);

	    context->SerialIO->IOSer.io_Command = SDCMD_QUERY;
	    r = DoIO ((struct IORequest *)context->SerialIO);
	    if (r)
		printf ("CM_input: DoIo 1 %d\n", r);

/*
 * Check for CD
 */
	    context->carrier_detect = !(context->SerialIO->io_Status & (1 << 5));
/*
 * if (context->SerialIO->IOSer.io_Actual & (1 << 5))
 */

	    if (CM_debug > 1)
		printf ("CM_check_CD: CD %d\n", context->carrier_detect);
	    CM_request_input (context);
	}
    }
}

/*
 * Cancel any outstanding input requests
 */
void CM_cancel_input (COMMS_CONTEXT * cm)
{
    if (cm->input_requested)
    {
	if (CheckIO ((struct IORequest *)cm->SerialIO))
	{
	    CM_input (cm, NULL);
	}
	AbortIO ((struct IORequest *)cm->SerialIO);
	WaitIO ((struct IORequest *)cm->SerialIO);
	cm->input_requested = FALSE;
	if (CM_debug)
	    printf ("CM_cancel_input: completed\n");
    }
    else
	printf ("CM_cancel_input: no request in progress\n");
}

/*
 * This routine actually opens the device.
 */
TY_int CM_open_device (COMMS_CONTEXT * cm, COMMS_CONFIG * config)
{
    TY_int result;

    if (cm->SerialMP = CreatePort (0, 0))
    {
	if (cm->SerialIO = (struct IOExtSer *)
	    CreateExtIO (cm->SerialMP, sizeof (struct IOExtSer)))
	{
	    cm->SerialIO->io_SerFlags = 0;	/*
						 * Example of setting flags 
						 */
/*
 * cm->SerialIO->IOSer.io_Message.mn_ReplyPort = cm->SerialMP;
 */

	    if (OpenDevice (config->settings->serialname, config->settings->unit,
			    (struct IORequest *)cm->SerialIO, 0))
	    {
		error_log ("Unable to open communications device %s",
			   config->settings->serialname);
		return (NULL);
	    }
	    else
	    {
/*
 * Now that we've opened the device, we will initially modify the
 * parameters to suit our requirements...
 */
		cm->device_opened = TRUE;
		cm->SerialIO->IOSer.io_Command = SDCMD_SETPARAMS;
/*
 * cm->SerialIO->io_SerFlags = ~SERF_PARTY_ON;
 */
		cm->SerialIO->io_SerFlags = SERF_XDISABLED | SERF_RAD_BOOGIE;

/*cm->SerialIO->io_SerFlags |= SERF_SHARED;*/

/*
 * strcpy (&cm->SerialIO->io_TermArray, "\r\n?");
 */

		if (config->settings->sevenwire)
		    cm->SerialIO->io_SerFlags |= SERF_7WIRE;

		cm->SerialIO->io_Baud = config->settings->baud;
		cm->SerialIO->io_ReadLen = 8;
		cm->SerialIO->io_WriteLen = 8;
		cm->SerialIO->io_StopBits = 1;
		cm->SerialIO->io_RBufLen = 16384;
		cm->SerialIO->IOSer.io_Command = SDCMD_SETPARAMS;
		DoIO ((struct IORequest *)cm->SerialIO);
#ifdef ZA
/*
 * io_ReadLen = IO_WriteLen = config->settings->length ? 7 : 8;
 * 
 * io_ExtFlags =  SEXTF_MSPON  /*  Enable parity 
 */
		SEXTF_MARK	/*
				 * For mark parity 
				 */
		    SERF_PARTY_ON
		    SERF_PARTY_ODD
		    config->settings->parity 0 = off, 1 = odd 2 = even

		    io_StopBits config->settings->bits 0 = 8 1 = 7
		    config->settings->xonflg ! SERF_XDISABLED
		    * /
#endif

		/*
		 * Precalculate a wait mask for the message port signals.
		 * NOTE: A signal may show up without an associated message!
		 */
		    cm->WaitMask = (1L << cm->SerialMP->mp_SigBit);

		if (result = DoIO ((struct IORequest *)cm->SerialIO))
		    printf ("Error setting parameters - code %ld!\n", result);

/*
 * Clone this read channel to some extent, to get the write channel.
 */
		cm->SerialWriteMP = CreatePort (0, 0);
		cm->SerialWriteIO = (struct IOExtSer *)
		    CreateExtIO (cm->SerialWriteMP, sizeof (struct IOExtSer));

		if (cm->SerialWriteMP && cm->SerialWriteIO)
		{

		    /*
		     * Copy over the entire old IO request, then stuff the
		     * new Message port pointer.
		     */
		    CopyMem (cm->SerialIO, cm->SerialWriteIO, sizeof (struct IOExtSer));
		    cm->SerialWriteIO->IOSer.io_Message.mn_ReplyPort = cm->SerialWriteMP;
/*
 * cm->SerialWriteIO->io_ExtFlags = cm->SerialIO->io_ExtFlags;
 * cm->SerialWriteIO->io_SerFlags = cm->SerialIO->io_SerFlags;
		     */ cm->SerialWriteIO->IOSer.io_Command = SDCMD_SETPARAMS;
		    DoIO ((struct IORequest *)cm->SerialWriteIO);

/*
 * Initiliase the circular buffer for use with whens.
 */
#ifdef USE_CB
		    if (!CB_init_buffer (&cm->cb, sizeof (char), 200))
		    {
			error_log ("CM_open: Cannot get circular buffer");
			return (NULL);
		    }
#endif
		    cm->first_wi = NULL;
		}

		cm->SerialIO->IOSer.io_Command = SDCMD_QUERY;
		DoIO ((struct IORequest *)cm->SerialIO);

		if (cm->SerialIO->IOSer.io_Actual)
		{
		    cm->SerialIO->IOSer.io_Command = CMD_READ;
		    cm->SerialIO->IOSer.io_Length = min (SERIAL_BUFFER_SIZE,
					     cm->SerialIO->IOSer.io_Actual);
		    cm->SerialIO->IOSer.io_Data = (APTR) cm->input_buffer;
		    DoIO ((struct IORequest *)cm->SerialIO);
		}
		return (TRUE);
	    }
	}
    }
    return (FALSE);
}

void CM_close_device(COMMS_CONTEXT *cm)
{
    if (cm->SerialWriteMP)
	DeletePort (cm->SerialWriteMP);
    if (cm->SerialWriteIO)
	DeleteExtIO ((struct IORequest *)cm->SerialWriteIO);

    if (cm->device_opened)
	CloseDevice ((struct IORequest *)cm->SerialIO);

    if (cm->SerialIO)
	DeleteExtIO ((struct IORequest *)cm->SerialIO);
    if (cm->SerialMP)
	DeletePort (cm->SerialMP);
}

TY_int CM_read_config (COMMS_CONTEXT * cm)
{
return NC_read_config(cm);
}

void CM_delay(TY_int ms)
{
Delay(ms/20);
}

