/*---------------------------------------------------------------------------
 *
 *      Title:          NICOLA - Resource Database List handling
 *
 *      File:           RE_lv.c
 *
 *      Date:           8 Aug 1993
 *
 *      Description:
 *              This module is responsible for ListView handling.
 *
 *      Copyright  1993 R.J.Harrison
 *
 *      Rev:    By              Done                                    Date
 *      1.00    R.J.Harrison    Implement                            8  June  92
 *      1.01    R.J.Harrison    Port to windows                      8  Aug   93
 *
 *---------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef AMIGA
#include <exec/types.h>
#include <exec/memory.h>
#include <exec/libraries.h>
#ifdef NO_INLINE
#include <clib/exec_protos.h>
#else
#include <proto/exec.h>
#endif
#endif

#include "zpl.h"

#ifdef TEST
char *decrud (char *a)
{
    return a;
}
TY_int filesize (FILE * f)
{
    return 12000;
}
FILE *create_message_save (char *d, TY_int a)
{
    return 0;
}
void message_save (FILE * f, char *i)
{
}
void finish_message_save (FILE * f, TY_int l)
{
}
#else
#endif


#ifndef AMIGA
void NewList (struct List *lis)
{
    lis->lh_Head = (struct Node *) &lis->lh_Tail;
    lis->lh_TailPred = (struct Node *) &lis->lh_Head;
    lis->lh_Tail = 0;
}

struct Node *
  AddTail (struct List *lis, struct Node *new_node)
{
    struct Node *enode = lis->lh_TailPred;

    enode->ln_Succ = new_node;
    new_node->ln_Succ = (struct Node *) &lis->lh_Tail;
    new_node->ln_Pred = enode;
    lis->lh_TailPred = new_node;
    return (new_node);
}

struct Node *
  RemTail (struct List *lis)
{
    struct Node *n = lis->lh_TailPred, *prev = n->ln_Pred;
    if (prev)
    {
	prev->ln_Succ = (struct Node *) &lis->lh_Tail;
	lis->lh_TailPred = prev;
	return (n);		/*
				 * Return the current end of list 
				 */
    }
    else
	return (0);
}
/*
 * We don't need, or use RemHead much...
 */
struct Node *
  RemHead (struct List *lis)
{
    return (RemTail (lis));
}

#endif

/*
 * This routine 'displays' a file into a list, making efficient use of
 * allocation.
 */

FILE_LIST *
  NG_display_text (char *text)
{
    if (text)
    {
	TY_int i;
	char *ptr, *lptr;
	FILE_LIST *fl = calloc (1, sizeof (FILE_LIST));

	fl->size = strlen (text);
	fl->file_data = text;

	if (fl->file_data)
	{
	    fl->file_data_end = fl->file_data + fl->size + 1;
	    fl->file_data = text;
	}
	else
	    return (0);

	ptr = fl->file_data;
	fl->lines = 0;

	while (ptr && (ptr < fl->file_data_end) && *ptr)
	{
	    ptr = strchr (ptr, '\n');
	    if (ptr)
	    {
		ptr++;
	    }
	    fl->lines++;
	}

	fl->nodes = calloc (fl->lines, sizeof (*fl->nodes));
	NewList (&fl->list);
	ptr = fl->file_data;
	i = 0;
	while (ptr && (ptr < fl->file_data_end) && *ptr)
	{
	    lptr = ptr;
	    ptr = strchr (ptr, '\n');
	    if (ptr)
	    {
		*ptr++ = 0;
	    }
	    fl->nodes[i].ln_Name = decrud (lptr);
	    AddTail (&fl->list, &fl->nodes[i]);
	    i++;
	}
	return (fl);
    }
    return (0);
}
FILE_LIST *
  NG_display_file (FILE * f)
{
    if (f)
    {
	char *ptr = make_text_from_open_file (f);
	fclose (f);
	return (NG_display_text (ptr));
    }
    return (0);
}

FILE_LIST *
  NG_display_named_file (char *filename)
{
    FILE *f = fopen (filename, "r");
    if (f)
	return (NG_display_file (f));
    return (NULL);
}

void
  NG_free_file (FILE_LIST * fl)
{
    if (fl)
    {
	if (fl->list.lh_Head)
	    while (RemTail (&fl->list)) ;
	free (fl->nodes);
	free (fl->file_data);
	free (fl);
    }
}

/*
 * Listview routines
 */
struct List *
  NG_initlv (struct List *lis)
{
    struct Node *node;

    if (lis->lh_Head)
	while ((node = RemHead (lis)) != NULL)
	    free (node);

    NewList (lis);
    return (lis);
}

void
  NG_freelv (struct List *lis)
{
    struct Node *node;

    if (lis->lh_Head)
	while ((node = RemTail (lis)) != NULL)
	    free (node);
}

void
  NG_freelv_and_data (struct List *lis)
{
    struct Node *node;

    if (lis->lh_Head)
	while ((node = RemTail (lis)) != NULL)
	{
	    free (node->ln_Name);
	    free (node);
	}
}

void
  NG_addlv (struct List *lis, void *item)
{
    struct Node *node;

    node = calloc (1, sizeof (struct Node));
    node->ln_Name = item;
    AddTail (lis, node);
}

/*
 * Counts the list size
 */

TY_int
NG_lv_size (struct List *lis)
{
    struct Node *node;
    TY_int count = 0;

    if (lis && lis->lh_Head)
    {
	node = lis->lh_Head;
	while (node->ln_Succ)
	{
	    count++;
	    node = node->ln_Succ;
	}
    }
    return (count);
}

/*
 * Ok, to sort a list (of pointers) into alpha-order we will 
 * basically setup an array of these pointers, then sort this
 * array, and then whizz through the list and modify the ln_Name
 * elements to be in the same order as the sorted array...
 */

static int cmp_lv_item (i1, i2)
char **i1, **i2;
{
    return (strcmp (*i1, *i2));
}

void NG_sort_list (struct List *lis, int (*cmp)(const void *, const void *))
{
    TY_int list_size = NG_lv_size (lis);
    if (list_size > 1)
    {
	char **tlist = calloc (sizeof (*tlist), list_size), **t;
	struct Node *node;

	if (!tlist)
	    return;

	t = tlist;
	node = lis->lh_Head;
	while (node->ln_Succ)
	{
	    *t = node->ln_Name;
	    t++;
	    node = node->ln_Succ;
	}
	qsort (tlist, list_size, sizeof (*tlist), cmp);

	t = tlist;
	node = lis->lh_Head;
	while (node->ln_Succ)
	{
	    node->ln_Name = *t;
	    t++;
	    node = node->ln_Succ;
	}
    }
}

void NG_alpha_sort_list (struct List *lis)
{
NG_sort_list(lis, cmp_lv_item);
}

/*
 * Returns the 'nth' element in the list
 */

void
  NG_list_output (FILE * f, struct List *lis)
{
    struct Node *node;
    if (f && lis && lis->lh_Head)
    {
	node = lis->lh_Head;
	while (node->ln_Succ)
	{
	    fprintf (f, "%s\n", node->ln_Name);
	    node = node->ln_Succ;
	}
	fprintf (f, "\n");
    }
}

void
  NG_print_list (struct List *lis, char *header)
{
    if (lis && lis->lh_Head)
    {
	FILE *f = create_message_save (NULL, TRUE);
	if (f)
	{
	    NG_list_output (f, lis);
	    finish_message_save (f, TRUE);
	}
    }
}

void NG_save_list (struct List *lis, char *filename)
{
    if (lis && lis->lh_Head)
    {
	FILE *f = create_message_save (filename, FALSE);
	if (f)
	{
	    NG_list_output (f, lis);
	    finish_message_save (f, FALSE);
	}
    }
}

void *
  NG_find_element (struct List *lis, TY_int element)
{
    struct Node *node;
    TY_int count = 0;

    if (lis && lis->lh_Head)
    {
	node = lis->lh_Head;
	while (node->ln_Succ)
	{
	    if (element == count++)
		return (node->ln_Name);
	    node = node->ln_Succ;
	}
    }
    return (NULL);
}

/*
 * Locates the first occurence of a number in a list
 */

TY_int
NG_find_int (struct List * lis, TY_int id, TY_int (*func) (void *, TY_int))
{
    struct Node *node;
    TY_int count = 0;

    if (lis && lis->lh_Head)
    {
	node = lis->lh_Head;
	while (node->ln_Succ)
	{
	    if ((*func) (node->ln_Name, id))
		return (count);
	    count++;
	    node = node->ln_Succ;
	}
    }
    return (-1);
}

/*
 * Locates where a certain string appears in a list
 */

TY_int
NG_find_name (struct List * lis, char *name)
{
    struct Node *node;
    TY_int count = 0;

    if (lis && lis->lh_Head)
    {
	node = lis->lh_Head;
	while (node->ln_Succ)
	{
	    if (!strcmp (node->ln_Name, name))
		return (count);
	    count++;
	    node = node->ln_Succ;
	}
    }
    return (-1);
}

TY_int
NG_match_string (struct List * lis, char *name, TY_int start,
		 TY_int (*func) (void *, char *, void *), void *ptr)
{
    struct Node *node;
    TY_int count = 0;

    if (lis && lis->lh_Head)
    {
	node = lis->lh_Head;
	while (node->ln_Succ)
	{
	    if (count > start)
	    {
		if ((*func) (node->ln_Name, name, ptr))
		    return (count);
	    }
	    count++;
	    node = node->ln_Succ;
	}
    }
    return (-1);
}

#ifdef TEST
main ()
{
    static struct List lv;
    NG_initlv (&lv);
    NG_addlv (&lv, "1Hair");
    NG_addlv (&lv, "2Hair Black");
    NG_addlv (&lv, "3Blue Eyes");
    NG_addlv (&lv, "4Brown Eyes(prevlast)");
    NG_addlv (&lv, "5Green Eyes(last)");
    RemTail (&lv);
    printf ("List Size %d\n", NG_lv_size (&lv));
    NG_list_output (stdout, &lv);
    NG_addlv (&lv, "New Last");
    NG_list_output (stdout, &lv);
    NG_freelv (&lv);
}

#endif
