/*---------------------------------------------------------------------------
 *
 *      Title:          NICOLA - Resource Database main module
 *
 *      File:           RE_main.c
 *
 *      Date:           8 June  1992
 *
 *      Description:
 *              This module provides the parser and general interface to
 *      the resource database
 *
 *      Copyright  1993 R.J.Harrison
 *
 *      Rev:    By              Done                                    Date
 *      1.00    R.J.Harrison    Implement                               8  June  92
 *      1.01    R.J.Harrison    Allow creation of an empty resource     23 Jan 93
 *      1.02    R.J.Harrison    Remove leading '.' on gadget text       07 June 93
 *
 *---------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>

#ifdef AMIGA
#include <intuition/intuition.h>
#include <intuition/gadgetclass.h>
#include <intuition/screens.h>
#include <libraries/gadtools.h>
#include <graphics/displayinfo.h>
#include <graphics/gfx.h>

#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/gadtools.h>
#include <pragmas/gadtools_pragmas.h>
#include <proto/graphics.h>
#include <proto/diskfont.h>

#endif

#ifdef AMIGA
#include "intuisup.h"
#endif

#include "zpl.h"

#ifdef SHORTC			// This is a fudge..... (or it would be if un#if'd)
#ifdef calloc
#undef calloc
#endif
#ifdef malloc
#undef malloc
#endif
#ifdef free
#undef free
#endif
#ifdef savestr
#undef savestr
#endif

#define calloc(a,b) _FW_calloc(fw_id,a,b,MODULE,__LINE__)
#define malloc(a) _FW_malloc(fw_id,a,MODULE,__LINE__)
#define free(a) _FW_free(fw_id,a,MODULE,__LINE__)
#define savestr(a) _FW_savestr(fw_id,a,MODULE,__LINE__)
#endif

#define SC_reset_offsets y_offset=x_offset=0;y_offset = SCFY (1) + 8;

#define NUM_GRAN 10
#define C_free(f,x) if(x) FW_free(f,x)

#ifdef AMIGA
extern TY_int using_v36 ();
extern struct Library *IntuiSupBase;
#endif

typedef struct
{
    TY_int xs, ys, x_offset, y_offset;
} SS;


typedef struct
{
    TY_int lessthan, compare_size, use_size;
} AUTO_SIZE;

static char *RE_database = NULL, *RE_database_end = NULL;
TY_int RE_allow_scale_y = TRUE;

static struct TextAttr *fixed_font;

/*
 * Count the size of the namelist given.
 */
static TY_int sizenamelist (char *ptr)
{
    char *p = ptr, *e;
    TY_int cur_offset = 0;

    while (p && *p)
    {
	p = skip_spaces (p);
	if (*p == '"')
	{
	    e = strchr (++p, '"');
	    if (e)
		*e++ = NULL;
	}
	else
	{
	    e = strchr (p, ',');
	    if (e)
		*e = NULL;
	}
	cur_offset++;

	if (e)
	    p = ++e;
	else
	    p = NULL;
    }
#ifdef DEBUG_VERSION
    if (re_debug)
	printf ("sizenamelist() = %d\n", cur_offset);
#endif

    if (!cur_offset)
    {
	error_log ("sizenamelist(): Error sizing namelist,seems empty");
	return (NULL);
    }
    else
	return (cur_offset);
}

/*
 * Split a string up into an array of pointers to the delimited parts.
 */
static char **get_name_list (RE_resource * rs, char *ptr)
{
    char *p = ptr, *e;
    char **names;
    TY_int num = 10, cur_offset = 0;

    if (!(names = FW_calloc (rs->window, num, sizeof (*names))))
	return (NULL);

    while (p && *p)
    {
	p = skip_spaces (p);
	if (*p == '"')
	{
	    e = strchr (++p, '"');
	    if (e)
		*e++ = NULL;
	}
	else
	{
	    e = strchr (p, ',');
	    if (e)
		*e = NULL;
	}
	names[cur_offset++] = FW_savestr (rs->window, p);

/*
 * When we run out of space in our current allocated lump, get
 * another chunk.
 */
	if (cur_offset >= num)
	{
	    TY_int old_num = num;
	    char **new_names;

	    num += NUM_GRAN;
	    if (!(new_names = FW_calloc (rs->window, num, sizeof (*names))))
	    {
		error_log ("get_name_list(): failed to get more memory for new");
		return (NULL);
	    }
	    memcpy (new_names, names, sizeof (*names) * old_num);
	    FW_free (rs->window, names);
            names=new_names;
	}
#ifdef DEBUG_VERSION
	if (re_debug)
	    printf ("-%s-\n", p);
#endif
	if (e)
	    p = ++e;
	else
	    p = NULL;
    }
#ifdef DEBUG_VERSION
    if (re_debug)
	printf ("get_name_list() = %x\n", names);
#endif
    if (!cur_offset)
    {
	error_log ("get_name_list(): Error creating namelist");
	FW_free (rs->window, names);
	return (NULL);
    }
    else
	return (names);
}

/*
 * Parse the given string, into the required parts, and create a new resource
 * structure for it. This new resource (gadget) structure is a pointer to a static
 * chunk, because the usage will only be to create a Gadget and a resource entry.
 */

static RE_new_gadget *getng (RE_resource * rs, char *ptr)
{
    static struct NewGadget ng;
    char *p1;
    static RE_new_gadget nr;
    static TY_int gadget_id = 0, lx = 1, ly = 1;

    if (!ptr)
	return (NULL);

    p1 = strtok (ptr, ",;/");
    if (!p1)
	return (NULL);

    memset (&ng, 0, sizeof (ng));
    memset (&nr, 0, sizeof (nr));

    nr.resource_id = FW_savestr (rs->window, p1);

    if (!(p1 = strtok (NULL, ",;/")))
	return (NULL);
    p1 = skip_spaces (p1);
    if (*p1 == '+')
	ng.ng_LeftEdge = ++lx;
    else if (*p1 == '=')
	ng.ng_LeftEdge = lx;
    else
	lx = ng.ng_LeftEdge = strtol (p1, NULL, 10);


    if (!(p1 = strtok (NULL, ",;/")))
	return (NULL);
    p1 = skip_spaces (p1);
    if (*p1 == '+')
	ng.ng_TopEdge = ++ly;
    else if (*p1 == '=')
	ng.ng_TopEdge = ly;
    else
	ly = ng.ng_TopEdge = strtol (p1, NULL, 10);

    if (!(p1 = strtok (NULL, ",;/")))
	return (NULL);
    ng.ng_Width = strtol (p1, NULL, 10);

    if (!(p1 = strtok (NULL, ",;/")))
	return (NULL);
    ng.ng_Height = strtol (p1, NULL, 10);

    if (!(p1 = strtok (NULL, ",;/")))
	return (NULL);
    ng.ng_Flags = strtol (p1, NULL, 10);

    if (!(p1 = strtok (NULL, ",;")))
	return (NULL);
    p1 = skip_spaces (p1);

    if (*p1 == '.' && p1[1] == ' ')
	*p1 = ' ';

    ng.ng_GadgetText = FW_savestr (rs->window, p1);
#ifdef WINDOWS
    gsub (ng.ng_GadgetText, '_', '&');
#endif
    if ((p1 = strtok (NULL, ",;/")))
	nr.size = strtol (p1, NULL, 10);
    else
	nr.size = 4;		/*
				 * Handy defaults 'r' us 
				 */

#ifdef ORS_1
    if (nr.size)
	nr.ptr = FW_calloc (rs->window, 1, nr.size);
#else
nr.ptr = 0;
#endif

    /*
     * Give this gadget a unique id. The gadget is initially identified by
     * its name, so the ID only has to be unique
     */
    nr.id = ng.ng_GadgetID = gadget_id++;
    nr.ng = &ng;
    nr.root = rs;
    return (&nr);
}

#ifdef DEBUG_VERSION
static void dump_gadget (RE_new_gadget * nr)
{
    MD_NEWGADGET_P ng = nr->ng;
    printf ("nr->resource_id %s\n", nr->resource_id);
    printf ("nr->ptr %d\n", nr->ptr);
    printf ("nr->size %d\n", nr->size);

    printf ("ng->ng_LeftEdge = %d\n", ng->ng_LeftEdge);
    printf ("ng->ng_TopEdge = %d\n", ng->ng_TopEdge);
    printf ("ng->ng_Width = %d\n", ng->ng_Width);
    printf ("ng->ng_Height = %d\n", ng->ng_Height);
    printf ("ng->ng_GadgetID = %d\n", ng->ng_GadgetID);
    printf ("ng->ng_Flags = %d\n", ng->ng_Flags);
    printf ("ng->ng_GadgetText = %s\n", ng->ng_GadgetText);

}
#endif

/*
 * Get a integer off the command line, and increment the pointer
 */
static TY_int get_int (char **p)
{
    TY_int i = strtol (*p, p, 10);
    return (i);
}

/*
 * This is the main line parsing routine. It will add the gadget onto the
 * list for this instance. Any post-processing of this gadget will be done
 * by locating it from the list, by the name.
 *
 * Input gadgets have in their structure a reference to the output variable.
 * By default this will be to dynamic memory allocated at creation, and will be
 * accessed by finding the name, and knowing the structure at the address.
 * - A calling routine would firstly load the gadget definitions.
 * - Setup a entity list using the loaded definitions.
 * - Save the entity list into the window structure, modify the window
 *   flags accordingly and return. The window then has the gadgets attached.
 */
static RE_new_gadget *parse (MD_GADGET_P g, char *line, RE_resource * r)
{
    char *ptr, *p1, *p2;
    MD_NEWGADGET_P ng;
    RE_new_gadget *nr;
    TY_int gex, gey;
    char oline[200];
    static AUTO_SIZE as[10];
    static TY_int asp = 0;

    strcpy (oline, line);

    if (!line || !*line)
	return (NULL);

#ifdef DEBUG_VERSION
    if (re_debug)
	printf ("parsing line /%s/\n", line);
#endif

    p1 = line;
    while (p1 = strchr (p1, '\t'))
	*p1 = ' ';

    p1 = strchr (line, '[');
    if (p1)
	p2 = strchr (p1, ']');
    else
	p2 = NULL;
    ptr = strtok (line, " ");

    if (!strcmp (ptr, "RO"))
    {
	SC_reset_offsets;
	return (NULL);
    }
    if (!strcmp (ptr, "NS"))
    {
	RE_allow_scale_y = FALSE;
	return (NULL);
    }
    if (!strcmp (ptr, "SC"))
    {
	RE_allow_scale_y = TRUE;
	return (NULL);
    }
/*
 * Auto Size the next gadget...
 */
    if (!strcmp (ptr, "AS"))
    {
	char *ptr1;

	asp = 0;
	memset (&as, 0, sizeof (as));
	while (ptr1 = strtok (NULL, ", /"))
	{
	    if (*ptr1 == '<')
	    {
		as[asp].lessthan = TRUE;
		*ptr1 = ' ';
	    }
	    if (*ptr1 == '>')
	    {
		as[asp].lessthan = FALSE;
		*ptr1 = ' ';
	    }
	    as[asp].compare_size = strtol (ptr1, NULL, 10);
	    ptr1 = strtok (NULL, ", /");
	    as[asp].use_size = strtol (ptr1, NULL, 10);
	    asp++;
	}

	return (NULL);
    }

    if (!p1 || !p2)
	return (NULL);

    if (p1[1] == '-')
    {
	p1++;
	RE_allow_scale_y = FALSE;
    }
    if (p1[1] == '+')
    {
	p1++;
	RE_allow_scale_y = TRUE;
    }

    *p2 = 0;
    p1++;
    if (!(nr = getng (r, p1)))
    {
	printf ("Error Parsing Line %s\n", oline);
	return (NULL);
    }
    if (asp)
    {
#ifdef AMIGA
	TY_int screen_height = r->scr->Height;
#else
	TY_int screen_height = FW_get_screen_height();
#endif
	TY_int size_to_use = 0, i;
	for (i = 0; i < asp && !size_to_use; i++)
	{
	    if (as[i].lessthan && screen_height < as[i].compare_size)
	    {
		size_to_use = as[i].use_size;
	    }
	    if (!as[i].lessthan && screen_height > as[i].compare_size)
	    {
		size_to_use = as[i].use_size;
	    }
	}
	asp = 0;
	if (size_to_use)
	    nr->ng->ng_Height = size_to_use;

#ifdef DEBUG_VERSION
	if (re_debug)
	    printf ("AS: used size %d, screen height %d\n", size_to_use, screen_height);
#endif
    }
#ifdef DEBUG_VERSION
    if (re_debug)
	dump_gadget (nr);
#endif
    ng = nr->ng;

    nr->window = r->window;
#ifdef AMIGA
    ng->ng_VisualInfo = r->vi;
    ng->ng_TextAttr = r->textattr;
#endif

#ifdef WINDOWS
    nr->instance = r->instance;
#endif

/*
 * if (!using_v36())
 */
    {
	if (strcmp (ptr, "LV"))
	    ng->ng_TopEdge--;
/*
 * ng->ng_LeftEdge--;
 */
	ng->ng_Width++;
    }
/*
 * Save the number of lines the gadget was created with, (LV scrolling)
 */
#ifdef WINDOWS
    if (!strcmp (ptr, "MX"))
	ng->ng_Width = strlen (nr->ng->ng_GadgetText) + 2;
    if (!strcmp (ptr, "BX"))
	ng->ng_Width = strlen (nr->ng->ng_GadgetText) + 2;
#endif
    nr->num_lines = ng->ng_Height - 3;
    nr->num_columns = ng->ng_Width - 3;
    ng->ng_LeftEdge = SCFX (ng->ng_LeftEdge);
    ng->ng_TopEdge = SCFY (ng->ng_TopEdge);
    ng->ng_Width = SCFX_rel (ng->ng_Width);

    if (ng->ng_Height < 50)
        ng->ng_Height = SCFY_rel (ng->ng_Height);

/*
 * CY [id,x,y,xs,ys,flags,text,size] namelist,
 * MX [x,y,xs,ys,flags,text] namelist,
 * BX [..] on,underscore
 * RO - Reset offsets after CY
 * LV [..] list,
 * NI [...] number
 * SI [...] string
 * BU [...] string
 */
    if (!strcmp (ptr, "CY"))
	g = NG_cycle (g, nr, get_name_list (r, p2 + 1));
    else if (!strcmp (ptr, "MX"))
	g = NG_mx (g, nr, get_name_list (r, p2 + 1), TRUE, 0);
    else if (!strcmp (ptr, "BX"))
	g = NG_box (g, nr, TRUE, FALSE);
    else if (!strcmp (ptr, "BU"))
	g = NG_button (g, nr);
/*
 * ~
 */
    else if (!strcmp (ptr, "CO"))
#ifdef AMIGA
	g = NG_palette (g, nr, r->scr->BitMap.Depth, SCFX (2));
#else
	g = NG_palette (g, nr, 8, SCFX (2));
#endif
    else if (!strcmp (ptr, "LV"))
    {
	p2++;
    ng->ng_TextAttr = fixed_font;
    if (nr->num_lines > 60)
        nr->num_lines /= FontY;
	g = NG_lv (g, nr, NULL, get_int (&p2));

    }
    else if (!strcmp (ptr, "NI"))
    {
	ng->ng_Height += 5;
	p2++;
	g = NG_numeric (g, nr, get_int (&p2), 0);
    }
    else if (!strcmp (ptr, "SI"))
    {
	ng->ng_Height += 5;
	p2++;
	g = NG_string (g, nr, get_int (&p2), nr->ptr);
    }
    else if (!strcmp (ptr, "VS"))
    {
	p2++;
	g = NG_sb (g, nr, get_int (&p2), 8, TRUE, TRUE);
    }
    else if (!strcmp (ptr, "HS"))
    {
	p2++;
	g = NG_sb (g, nr, get_int (&p2), 8, FALSE, TRUE);
    }
    else if (!strcmp (ptr, "SL"))
    {
	p2++;
	g = NG_slider (g, nr, get_int (&p2), get_int (&p2));
    }
    else if (!strcmp (ptr, "TE"))
    {
	p2++;
	g = NG_text (g, nr, get_int (&p2));
    }
    else
    {
	error_log ("resource_parse: Illegal token /%s/", ptr);
	return (NULL);
    }
    nr->g = g;
#define NGs
#ifdef NGs
    {
	gex = nr->ng->ng_LeftEdge + nr->ng->ng_Width;
	gey = nr->ng->ng_TopEdge + nr->ng->ng_Height;
    }
#else
    gex = g->LeftEdge + g->Width + x_offset;
    gey = g->Height + g->TopEdge + y_offset;
#endif

    if (nr->ng->ng_Flags & PLACETEXT_RIGHT)
	gex += SCFX_rel (1 + strlen (ng->ng_GadgetText));
    if (gex > r->max_x)
	r->max_x = gex;
    if (gey > r->max_y)
	r->max_y = gey;

    if (!g)
	error_log ("RE_parse(): Gadget creation failed %s", line);
    return (nr);
}


/*
 * This is called during startup and reads in the main resource file.
 * There is only one file, which is divided up into individual resource sections.
 */
TY_int RE_init_resource (char *filename)
{
    FILE *f;
    TY_int size;
#ifdef WINDOWS
    if ((f = fopen (filename, "R")))
#else
    if ((f = fopen (filename, "r")))
#endif
    {
	if (RE_database)
	    free (RE_database);

	size = filesize (f);
	RE_database = calloc (size + 1, sizeof (char));
	if (RE_database)
	{
            extern char re_database_version;

	    RE_database_end = RE_database + size + 1;
	    if (!fread (RE_database, sizeof (char), size, f))
	    {
		free (RE_database);
		RE_database = NULL;
		fclose (f);
		return (FALSE);
	    }
	    fclose (f);
	    if (RE_database[0] == '#' &&
		RE_database[1] == '!' &&
		RE_database[2] >= re_database_version)
	    {
		return (TRUE);
	    }
	    else
	    {
char aa[2];
aa[0]=RE_database[2];
aa[1]=0;
		tell_em ("Incorrect version (V1.%s) of resource database found\nPlease fix this before continuing", aa);
		free (RE_database);
		RE_database = NULL;
		return (FALSE);
	    }
	}
	return (TRUE);
    }
    return (FALSE);
}

static char *rgetline (char **ptr)
{
    STATIC_VAR(line,char,200);

    char *l = strchr (*ptr, '\n');
    if (!l)
	return (NULL);
    memset (line, 0, SV_SIZEOF(line));
    strncpy (line, *ptr, l - *ptr);
    *ptr = ++l;
    if (*line == '#')
	*line = 0;
    return (line);
}

static char *find_resource (char *name)
{
    char *ptr, *line;
int len = strlen(name);

    if (!RE_database)
	return (NULL);

    ptr = RE_database;
    while (ptr && (ptr < RE_database_end) && *ptr)
    {
	line = strchr(ptr,'%');
	if(line)
	{
	ptr = &line[1];
	if ( (line[-1] == '\n') && (ptr[len] == '\n') )
	{
	    if (!strncmp (name,ptr,len))
	    {
		    return (ptr);
	    }
	}
        }
        else
            ptr = 0;
    }
/*
 * Resource not found
 */
    return (NULL);
}

/*
 * Given a new resource structure, allocate and create a resource database
 * entry.
 */
RE_resource_entry *create_resource (RE_new_gadget * nr)
{
    RE_resource_entry *re;

    if (!nr)
    {
	error_log ("RE_resource_entry: NULL new_gadget request initiated");
	return (NULL);
    }
    re = FW_calloc (nr->root->window, 1, sizeof (*re));

    re->name = nr->resource_id;
    re->g = nr->g;

#ifdef WINDOWS
if(nr->type == INTEGER_KIND)
{
re->read = FW_calloc(nr->root->window,1,25);
re->write = nr->ptr;
}
else
{
    re->read = nr->ptr;
    re->write = re->read;
}
#else
    re->read = nr->ptr;
    re->write = re->read;
#endif
    re->last_value = FW_calloc (nr->root->window, 1, nr->size);
    re->next = NULL;
    re->item_size = nr->size;
    re->type = nr->type;
    re->key_shortcut = tolower (nr->key_shortcut);
    re->redisplay = FALSE;
    re->num_lines = nr->num_lines;
    re->num_columns = nr->num_columns;
    re->cur_pos = -1;
    re->vertical = nr->vertical;
    re->id = nr->id;
    re->root = nr->root;
    return (re);
}



/*
 * Scan for the resource given, and create a resource structure for it. This can be
 * called as often as required, for each window of this type (usually).
 *
 * It is the resposibility of the window opening routine to open a window of the correct size,
 * and pass a pointer to here.
 * The routine RE_get_sizes may be used to find the X and Y size to use.
 */
RE_resource *RE_create_resources (ZPL_WINDOW * fw, char *resource_name)
{
    char *reptr, *ptr;
    MD_GADGET_P g;
    RE_new_gadget *nr;
    RE_resource *r;
    RE_resource_entry *re, *last_re;
    TY_int done;
    static struct TextAttr *rf, *ffa;
    static struct TextAttr default_fixed = {"topaz.font",8,0,0};

    if (!RE_database)
	return (NULL);

    if (resource_name)
    {
	reptr = find_resource (resource_name);
	if (!reptr)
	    return (NULL);
    }

    r = FW_calloc (fw, 1, sizeof (*r));

    r->name = FW_savestr (fw, resource_name);

    ffa = PR_get_fw_font ();

    if (ffa && ffa->ta_Name && *ffa->ta_Name)
    {
	fixed_font = ffa;
    }
    else
        fixed_font = &default_fixed;


    rf = PR_get_re_font ();
    if (rf && rf->ta_Name && *rf->ta_Name)
    {
	r->textattr = rf;
    }
#ifdef AMIGA
    else
	r->textattr = get_window_font ();
    r->scr = get_screen ();


    /*
     * open the font and set the scaling factors.
     */

    y_offset = 8;
    x_offset = 2;
/*
 * Initialise the two fonts. The last font initialised will be the font used
 * to set the scaling.
 */

fixed_font = init_font (fixed_font,0);
r->textattr = init_font (r->textattr,0);

#else
    SCset (FW_get_font_x_size(FW_get_dialog_font()), FW_get_font_y_size(FW_get_dialog_font()));
    SC_reset_offsets;
#endif

    RE_allow_scale_y = TRUE;

#ifdef AMIGA
    if (using_v36 ())
    {
	if (!(r->vi = GetVisualInfo (r->scr, TAG_DONE)))
	{
	    error_log ("RE_create_resource(): Unable to get visual info");
	    return (NULL);
	}

	/*
	 * Start defining the gadgets...
	 */
	if (NOT (g = CreateContext (&r->G_list)))
	{
	    error_log ("RE_create_resource(): Unable to CreateContext for list %x", &r->G_list);
	    return (NULL);
	}
    }
    else
    {
	r->ri = IGetRenderInfo (NULL, NULL);	/*
						 * (USHORT)(RENDER_INFO_FLAG_I
						 * * NNER_WINDOW)); 
						 */
    }
#endif
#ifdef WINDOWS
    {
	r->instance = get_instance ();
    }
#endif
    r->window = fw;

    reptr = strchr (reptr, '\n');
    if (reptr)
	reptr++;
    done = FALSE;
    last_re = 0;		/*
				 * Mainly for the compiler... 
				 */

    if (resource_name)
    {
	while (reptr && *reptr && !done)
	{
	    nr = parse (g, (ptr = rgetline (&reptr)), r);
	    if (!ptr)
		done = TRUE;
	    else
	    {
		if (nr && nr->g)
		{
		    re = create_resource (nr);
//    if (re->type == LISTVIEW_KIND)
//	    RE_select_lv (re, 1);
#ifdef AMIGA
		    if (using_v36 ())
		    {
			g = nr->g;
			g->UserData = re;
		    }
#endif
		    if (r->first_resource)
			last_re->next = re;
		    else
			r->first_resource = re;
		    last_re = re;

		}
		if (*reptr == '%')
		    done = TRUE;
	    }
	}
	r->max_x += 13;
#ifdef AMIGA
	if (using_v36 ())
	    r->max_y -= get_window_title_height () - 2;
	else
#endif
	    r->max_y += 6;
	r->last_g = g;
    }
    return (r);
}

/*
 * Release memory used by a resource which is no longer in use
 */
TY_int RE_dispose (RE_resource * r)
{
#ifdef AMIGA
if(r)
{
    if (using_v36 ())
    {
if(r->G_list)
	FreeGadgets (r->G_list);
if(r->vi)
	FreeVisualInfo (r->vi);
    }
    else
    {
if(r->gl)
{
	IRemoveGadgets (r->gl);
	IFreeGadgets (r->gl);
}
    }
}
#endif
    return (0);
}

void RE_add_resources (RE_resource * r, MD_WINDOW_P win)
{
    TY_int count = 0;

    if (r && win)
    {
	r->win = win;
/*
 * Resize the window if it isn't big enough for the resources which
 * have just been added.
 */
#ifdef RESIZE_DURING_ADD
	dy = (r->max_y + 6	/*
				 * +win->BorderTop 
	       */ ) - win->Height;
	dx = (r->max_x + 6 + win->BorderLeft + win->BorderRight) - win->Width;

	if (abs (dx) > 3 || abs (dy) > 3)
	{
	    SizeWindow (win, dx, dy);
	    Delay (6);
	}
#endif

#ifdef AMIGA
	if (using_v36 ())
	{
	    /*
	     * Add gadgets to the window, refresh them, and then call the toolkit
	     * refresh function.
	     */
	    AddGList (win, r->G_list, ((UWORD) - 1), ((UWORD) - 1), NULL);
	    RefreshGList (r->G_list, win, NULL, ((UWORD) - 1));
	    GT_RefreshWindow (win, NULL);
	}
	else
	{

	    /*
	     * With IntuiSup it is a little more complex. IntuiSup requires an
	     * array rather than a linked list of gadgets, so create one from the
	     * resource list, and FW_free the 'new' gadgets as each one is added
	     * into the array
	     */
	    struct GadgetData *gadgets;
	    RE_resource_entry *re;
	    TY_int i;

	    re = r->first_resource;
	    count = 0;
	    while (re)
	    {
		count++;
		re = re->next;
	    }
	    gadgets = FW_calloc (r->window, (++count) + 1, sizeof (struct GadgetData));
	    if (gadgets)
	    {

		/*
		 * Make a list from the resource list containing all of the data
		 * just setup. Free the re->g entries as added onto here
		 */
		r->gadgets = gadgets;
		re = r->first_resource;
		for (i = 0; i < count - 1; i++)
		{
		    memcpy (&gadgets[i], re->g, sizeof (*gadgets));
		    FW_free (r->window, re->g);
		    re->g = (MD_GADGET_P) & gadgets[i];
		    re->id = i;
		    re = re->next;
		}
		gadgets[count - 1].gd_Type = INTUISUP_DATA_END;

		if (r->gl = ICreateGadgets (r->ri, gadgets, 0, 0, 0))
		    IDisplayGadgets (win, r->gl);
	    }
	}
#endif
#ifdef WINDOWS
	{
	    RE_resource_entry *re;
	    TY_int i;

	    re = r->first_resource;
	    while (re)
	    {
		ShowWindow (re->g, SW_SHOWNA);
if(re->type == LISTVIEW_KIND)
SendMessage(re->g, WM_SETFONT,
                    (WPARAM) FW_get_fixed_font(), 0);
else
SendMessage(re->g, WM_SETFONT,
                    (WPARAM) FW_get_dialog_font(), 0);
		re = re->next;
	    }
	}
#endif
    }
}

TY_int RE_set_scrollbar (RE_resource_entry * re, TY_int top, TY_int total, TY_int win_size)
{
    if (re)
    {
#ifdef AMIGA
	if (using_v36 ())
	{
	    if (win_size >= 0)
		GT_SetGadgetAttrs (re->g, re->root->win, NULL,
				   GTSC_Visible, win_size,
				   TAG_DONE);
	    if (top >= 0)
		GT_SetGadgetAttrs (re->g, re->root->win, NULL,
				   GTSC_Top, top,
				   TAG_DONE);
	    if (total >= 0)
		GT_SetGadgetAttrs (re->g, re->root->win, NULL,
				   GTSC_Total, total,
				   TAG_DONE);
	}
	else
	{
	    if (win_size < 0)
		win_size = USE_CURRENT_VALUE;
	    if (total < 0)
		total = USE_CURRENT_VALUE;
	    if (top < 0)
		top = USE_CURRENT_VALUE;
	    ISetGadgetAttributes (re->root->gl, find_gid (re),
				  NULL, NULL, win_size, total, (void*)top);
	}
#endif
    }
    return (1);
}

#ifdef AMIGA
static void new_scrollbar (MD_WINDOW_P win, MD_GADGET_P g, RE_resource * r)
{
    RE_resource_entry *re, *last_re;
    static TY_int temp;
    static struct NewGadget ng;
    static RE_new_gadget nr;

    ng.ng_LeftEdge = win->Width - win->BorderRight;
    ng.ng_TopEdge = win->BorderTop;
    ng.ng_Width = win->BorderRight;
    ng.ng_Height = win->Height - win->BorderTop - win->BorderBottom - 7;
    ng.ng_VisualInfo = r->vi;
    nr.ng = &ng;
    nr.g = NG_sb (g, &nr, FALSE, 8, TRUE, TRUE);
    nr.size = sizeof (TY_int);
    nr.ptr = &temp;
    nr.root = r;
    re = create_resource (&nr);
    re->name = "scrollbar";

    if (using_v36 ())
    {
	g = nr.g;
	g->UserData = re;
    }
    else
	ng.ng_LeftEdge = win->Width - win->BorderRight - 10;

    if (r->first_resource)
    {
	last_re = r->first_resource;
	while (last_re->next)
	    last_re = last_re->next;
	last_re->next = re;
    }
    else
	r->first_resource = re;

}
#endif

RE_resource *RE_add_scrollbar (RE_resource * r, MD_WINDOW_P win, ZPL_WINDOW *fw)
{
#ifdef AMIGA
    static char *resource_name = "Scrollbar";
    MD_GADGET_P g;

    if (!r)
    {
	r = FW_calloc (fw, 1, sizeof (*r));

	r->name = FW_savestr (fw, resource_name);
	r->scr = get_screen ();
r->window = fw;

	if (using_v36 ())
	{
	    if (!(r->vi = GetVisualInfo (r->scr, TAG_DONE)))
	    {
		error_log ("RE_create_resource(): Unable to get visual info");
		return (NULL);
	    }

	    /*
	     * Start defining the gadgets...
	     */
	    if (NOT (g = CreateContext (&r->G_list)))
	    {
		error_log ("RE_add_scrollbar(): Unable to CreateContext for list %x", &r->G_list);
		return (NULL);
	    }
	}
	else
	{
	    r->ri = IGetRenderInfo (NULL, NULL);	/*
							 * (USHORT)(RENDER_INFO_FLAG_I
							 * * NNER_WINDOW)); 
							 */
	}
    }
    else
	g = r->last_g;

    new_scrollbar (win, g, r);
    return (r);
#else
    return (0);
#endif
}

void free_resource_entry (RE_resource_entry * re)
{
    C_free (re->root->window, re->name);
/*
 * ~free gadget space ? re->g = nr->g;
 */
}
/*
 * Frees an entiry resource structure, and the data items associated with it.
 * Call this after any 'update_lists' (or anything else using the data items)
 * have been destroyed
 */
TY_int RE_free_resource (RE_resource * resource)
{
    RE_resource_entry *re, *nre;
    if (!resource)
    {
	error_log ("RE_free_resource(): Given NULL resource to free");
	return (FALSE);
    }
    re = resource->first_resource;
    if (!re)
	error_log ("RE_free_resource(): No resource entries to free");
    while (re)
    {
	nre = re->next;
	free_resource_entry (re);
	re = nre;
    }
    return (TRUE);
}


void RE_delete_resources (RE_resource * r, MD_WINDOW_P win)
{
#ifdef AMIGA
    FW_free (r->window, r);
#endif
}

void RE_change_gadget_text (RE_resource_entry *re, char *newtext)
{
#ifdef CGT_DONE
RE_resource *r = re->root;

	if (using_v36 ())
	{
struct IntuiText *new_itext = Setup_text (r->window, newtext);
	    RemoveGList (r->win, r->G_list, ((UWORD) - 1));
new_itext->FrontPen = re->g->GadgetText->FrontPen;
new_itext->BackPen = re->g->GadgetText->BackPen;
new_itext->LeftEdge = re->g->GadgetText->LeftEdge;
new_itext->TopEdge = re->g->GadgetText->TopEdge;
re->g->GadgetText = new_itext;
	    RE_add_resources (r, r->win);
	}
	else
	{
tell_em("can't do this with intuisup");
	}
#endif
}

RE_resource *RE_resize_scrollbar (RE_resource * r, ZPL_WINDOW *fw)
{
#ifdef AMIGA
    MD_WINDOW_P win = r->win;
    RE_resource_entry *re_sb;
    if (re_sb = RE_find_resource ("scrollbar", r))
    {
	if (using_v36 ())
	{
	    RemoveGList (win, r->G_list, ((UWORD) - 1));
	    RE_delete_resources (r, win);
	    r = RE_add_scrollbar (NULL, win, fw);
	    RE_add_resources (r, win);
	}
	else
	{
	    IModifyGadget (r->gl, find_gid (re_sb),
			   win->Width - win->BorderRight,
			   win->BorderTop,
			   win->BorderRight,
		      win->Height - win->BorderTop - win->BorderBottom - 7);
	    IRefreshGadgets (r->gl);
	}
    }
    else
	error_log ("Cannot find scrollbar to resize");
    return (r);
#else
    return (0);
#endif
}


#ifdef ALLOW_SIZING
/*
 * If this item makes the window bigger, then set the
 * maximum sizes to this..
 */
static RE_max_size (char *line, SS * ss)
{
    char *ptr, *p1, *p2;
    RE_new_gadget *nr;
    TY_int gex, gey;

    if (!line || !*line)
	return (NULL);

#ifdef DEBUG_VERSION
    if (re_debug)
	printf ("sizing line /%s/\n", line);
#endif

    p1 = strchr (line, '[');
    p2 = strchr (p1, ']');
    ptr = strtok (line, " ");

    if (!strcmp (ptr, "RO"))
    {
	SC_reset_offsets;
	return (NULL);
    }
    if (!strcmp (ptr, "NS"))
    {
	RE_allow_scale_y = FALSE;
	return (NULL);
    }
    if (!strcmp (ptr, "SC"))
    {
	RE_allow_scale_y = TRUE;
	return (NULL);
    }

    if (!p1 || !p2)
	return (NULL);

    if (p1[1] == '-')
    {
	p1++;
	RE_allow_scale_y = FALSE;
    }
    if (p1[1] == '+')
    {
	p1++;
	RE_allow_scale_y = TRUE;
    }

    *p2 = 0;
    nr = getng (p1 + 1);

    if (!strcmp (ptr, "MX"))
    {
	ss->y_offset += sizenamelist (p2 + 1);
    }

    gex = nr->ng->ng_LeftEdge + nr->ng->ng_Width + ss->x_offset;
    gey = nr->ng->ng_TopEdge + nr->ng->ng_Height + ss->y_offset;

    if (gex > ss->xs)
	ss->xs = gex;
    if (gey > ss->ys)
	ss->ys = gey;

}

TY_int RE_size_resource (char *resource_name, TY_int * xs, TY_int * ys)
{
    char *reptr, *ptr;
    TY_int done;
    SS sizing;

    return (0);
    memset (&sizing, 0, sizeof (sizing));

    *xs = 0;
    *ys = 0;

    if (!RE_database)
	return (FALSE);

    y_offset = 0;
    x_offset = 2;		/*
				 * win->BorderLeft; 
				 */

    SCset (1, 1);
    reptr = find_resource (resource_name);
    if (!reptr)
	return (FALSE);

    reptr = strchr (reptr, '\n');
    if (reptr)
	reptr++;
    done = FALSE;
    while (reptr && *reptr && !done)
    {
	RE_max_size ((ptr = rgetline (&reptr)), &sizing);
	if (!ptr || *reptr == '%')
	    done = TRUE;
    }
    *xs = sizing.xs;
    *ys = sizing.ys;
    return (TRUE);
}

#endif



/*
 * From a resource structure, find the given resource entry. This is usually used to
 * find the root Glist, or to get the individual data item adresseses.
 * --exact matches only, please--
 */
RE_resource_entry *RE_find_resource (char *name, RE_resource * resource)
{
    RE_resource_entry *re;
    if (!resource)
    {
	error_log ("RE_find_resource(): Given NULL resource to search");
	return (FALSE);
    }
    re = resource->first_resource;
    if (!re)
	error_log ("RE_find_resource(): No resource entries");
    while (re)
    {
	if (!strcmp (name, re->name))
	    return (re);
	re = re->next;
    }
    return (NULL);
}

RE_resource_entry *RE_locate_resource_id (int id, RE_resource * resource)
{
    RE_resource_entry *re;
    if (!resource)
    {
	error_log ("RE_find_resource(): Given NULL resource to search");
	return (FALSE);
    }
    re = resource->first_resource;
    if (!re)
	error_log ("RE_find_resource(): No resource entries");
    while (re)
    {
	if (re->id == id)
	    return (re);
	re = re->next;
    }
    return (NULL);
}

void RE_set_process (RE_resource_entry * re, void (*process) (), void *param)
{
    if (re)
    {
	re->process = process;
	re->process_param = param;
    }
}

/*
 * Function to handle a GADGETUP or GADGETDOWN event.  For toolkit gadgets,
 * it is possible to use this function to handle MOUSEMOVEs as well, with
 * little or no work.
 *
 */
#define get_gd(a) (struct GadgetData *)(a)
/*
 * As intuisup needs a poxy offset (what's wrong with just an address FFS),
 * for the gadget, we'll have to whizz through the list until we find the
 * correct id.
 */
TY_int find_gid (RE_resource_entry * re)
{
    return (re->id);
}


void RE_add_lv (RE_resource_entry * re, void *ptr)
{
    re->size = NG_lv_size (ptr);
    if (re->size < re->cur_pos)
	re->cur_pos = re->size;

#ifdef AMIGA
    if (using_v36 ())
	GT_SetGadgetAttrs (re->g, re->root->win, NULL,
			   GTLV_Labels, ptr, TAG_DONE);
    else
    {
	ISetGadgetAttributes (re->root->gl, find_gid (re),
			      NULL, NULL, USE_CURRENT_VALUE, 1, ptr);
	RE_select_lv (re, 0);
    }
#endif
#ifdef WINDOWS
    {
	struct Node *node;
	struct List *lis = (struct List *) ptr;
	SendMessage (re->g, LB_RESETCONTENT, 0, 0);
	if (lis && lis->lh_Head)
	{
	    node = lis->lh_Head;
	    while (node->ln_Succ)
	    {
		SendMessage (re->g, LB_ADDSTRING, 0, (LPARAM) node->ln_Name);
		node = node->ln_Succ;
	    }
	}
if(re->cur_pos <= re->size)
	RE_select_lv (re, re->cur_pos);
else
	RE_select_lv (re, 0);
    }
#endif
    re->read = ptr;
}

void RE_select_lv (RE_resource_entry * re, TY_int index)
{
#ifdef AMIGA
    if (using_v36 ())
    {
	GT_SetGadgetAttrs (re->g, re->root->win, NULL,
			   GTLV_Selected, index, TAG_DONE);
	RE_set_value (re, index);
    }
    else
    {
	ISetGadgetAttributes (re->root->gl, find_gid (re),
	   NULL, NULL, USE_CURRENT_VALUE, (ULONG) index, (void*)USE_CURRENT_VALUE);
    }
#endif
#ifdef WINDOWS
    SendMessage (re->g, LB_SETCURSEL, index, 0);
#endif
    re->cur_pos = index;
    RE_set_value (re, re->cur_pos);
}

void RE_set_top_lv (RE_resource_entry * re, TY_int index)
{
    re->cur_pos = index;
#ifdef AMIGA
    if (using_v36 ())
	GT_SetGadgetAttrs (re->g, re->root->win, NULL,
			   GTLV_Top, index, TAG_DONE);
    else
    {
	ISetGadgetAttributes (re->root->gl, find_gid (re),
	   NULL, NULL, USE_CURRENT_VALUE, (ULONG) index, (void*)USE_CURRENT_VALUE);
    }
#endif
#ifdef WINDOWS
//SendMessage(re->g, LB_SETCURSEL, index, 0);
#endif
}

void RE_scroll_lv (RE_resource_entry * re, TY_int up)
{
    if (re->type == LISTVIEW_KIND)
    {
#ifdef USING_V39
    if (get_gadtools_version() >= 39)
{
GT_GetGadgetAttrs (re->g, re->root->win, NULL, GTLV_Selected, &re->cur_pos, TAG_DONE);
}
#endif
	if (up && re->cur_pos > 0)
	{
	    RE_set_top_lv (re, re->cur_pos - 1);
	    RE_select_lv (re, re->cur_pos);
	}
	else if (!up && re->cur_pos < (re->size - 1))
	{
	    RE_set_top_lv (re, re->cur_pos + 1);
	    RE_select_lv (re, re->cur_pos);
	}
    }
}

void RE_page_scroll_lv (RE_resource_entry * re, TY_int up)
{
    if (re->type == LISTVIEW_KIND)
    {
	TY_int nl = re->num_lines - 1;
	if (up)
	{
	    if (re->cur_pos - nl >= 0)
		RE_set_top_lv (re, re->cur_pos - nl);
	    else
		RE_set_top_lv (re, 0);
	    RE_select_lv (re, re->cur_pos);
	}
	else
	{
	    if (re->cur_pos < re->size - nl)
		RE_set_top_lv (re, re->cur_pos + nl);
	    else
		RE_set_top_lv (re, re->size - 1);
	    RE_select_lv (re, re->cur_pos);
	}
    }
}

void *RE_detach_lv (RE_resource_entry * re)
{
#ifdef AMIGA
    if (using_v36 ())
	GT_SetGadgetAttrs (re->g, re->root->win, NULL,
			   GTLV_Labels, ~0, TAG_DONE);
    else
    {
	ISetGadgetAttributes (re->root->gl, find_gid (re),
	      NULL, NULL, USE_CURRENT_VALUE, USE_CURRENT_VALUE, (VOID *) 0);
    }
#endif
#ifdef WINDOWS
#endif
    return (re->read);
}

void *RE_remove_lv (RE_resource_entry * re)
{
#ifdef AMIGA
    if (using_v36 ())
	GT_SetGadgetAttrs (re->g, re->root->win, NULL,
			   GTLV_Labels, NULL, TAG_DONE);
    else
    {
	ISetGadgetAttributes (re->root->gl, find_gid (re),
	      NULL, NULL, USE_CURRENT_VALUE, USE_CURRENT_VALUE, (VOID *) 0);
    }
#endif
#ifdef WINDOWS
    SendMessage (re->g, LB_RESETCONTENT, 0, 0);
#endif
    return (re->read);
}

void RE_reattach_lv (RE_resource_entry * re)
{
#ifdef AMIGA
    if (using_v36 ())
	GT_SetGadgetAttrs (re->g, re->root->win, NULL,
			   GTLV_Labels, re->read, TAG_DONE);

    else
    {
	ISetGadgetAttributes (re->root->gl, find_gid (re),
	NULL, NULL, USE_CURRENT_VALUE, USE_CURRENT_VALUE, (void *) re->read);
    }
#endif
}

TY_int RE_get_value (RE_resource_entry * re)
{
if(!re->write)
return(0);
    switch (re->item_size)
    {
    case 1:
	return (TY_int) (*(char *) re->write);
    case 2:
	return (TY_int) (*(unsigned short *) re->write);
    default:
	return (*(TY_int *) re->write);
    }
}

TY_int RE_get_named_value (char *name, RE_resource * r)
{
    RE_resource_entry *re = RE_find_resource (name, r);
    if (re)
    {
	return (RE_get_value (re));
    }
    return (0);
}

void RE_set_value (RE_resource_entry * re, TY_int value)
{
if(!re->write)
return;
    switch (re->item_size)
    {
    case 1:
	(*(char *) re->write) = (char) value;
	break;
    case 2:
	(*(unsigned short *) re->write) = (unsigned short) value;
	break;
    default:
	(*(TY_int *) re->write) = value;
    }
}

void RE_set_named_value (char *name, TY_int value, RE_resource * r)
{
    RE_resource_entry *re = RE_find_resource (name, r);
    if (re)
    {
	RE_set_value (re, value);
    }
}

void RE_update_item (RE_resource_entry * re, TY_int force)
{
    unsigned TY_int *iptr = re->write;

    if (re->update_in_progress || !iptr)
	return;

    if (force || memcmp (re->write, re->last_value, re->item_size))
    {
	re->update_in_progress = 1;
#ifdef AMIGA
	if (using_v36 ())
	{
	    switch (re->type)
	    {
	    case INTEGER_KIND:
		GT_SetGadgetAttrs (re->g, re->root->win, NULL,
				   GTIN_Number, RE_get_value (re), TAG_DONE);
		break;

	    case STRING_KIND:
		GT_SetGadgetAttrs (re->g, re->root->win, NULL,
				   GTST_String, re->read, TAG_DONE);
		break;

	    case TEXT_KIND:
		GT_SetGadgetAttrs (re->g, re->root->win, NULL,
				   GTTX_Text, re->read, TAG_DONE);
		break;

	    case CHECKBOX_KIND:
		GT_SetGadgetAttrs (re->g, re->root->win, NULL,
				 GTCB_Checked, RE_get_value (re), TAG_DONE);
		break;

	    case CYCLE_KIND:
		GT_SetGadgetAttrs (re->g, re->root->win, NULL,
				   GTCY_Active, RE_get_value (re), TAG_DONE);
		break;

	    case MX_KIND:
		GT_SetGadgetAttrs (re->g, re->root->win, NULL,
				   GTMX_Active, RE_get_value (re), TAG_DONE);
		break;

	    case SLIDER_KIND:
		GT_SetGadgetAttrs (re->g, re->root->win, NULL,
				   GTSL_Level, RE_get_value (re), TAG_DONE);
		break;

	    case SCROLLER_KIND:
		break;

	    case PALETTE_KIND:
		GT_SetGadgetAttrs (re->g, re->root->win, NULL,
				   GTPA_Color, RE_get_value (re), TAG_DONE);
		break;

	    case LISTVIEW_KIND:
		break;
	    }
	}
	else
	{
	    switch (re->type)
	    {
	    case SLIDER_KIND:
	    case SCROLLER_KIND:
/*
 * min,max,current
 */
		break;
	    case CHECKBOX_KIND:
		ISetGadgetAttributes (re->root->gl, find_gid (re),
				      NULL, NULL, RE_get_value(re), USE_CURRENT_VALUE, (VOID *) USE_CURRENT_VALUE);
		break;
	    case CYCLE_KIND:
	    case MX_KIND:
		ISetGadgetAttributes (re->root->gl, find_gid (re),
				      NULL, NULL, USE_CURRENT_VALUE, RE_get_value(re), (VOID *) USE_CURRENT_VALUE);
		break;

	    case PALETTE_KIND:
	    case INTEGER_KIND:
		ISetGadgetAttributes (re->root->gl, find_gid (re),
				      NULL, NULL, USE_CURRENT_VALUE, USE_CURRENT_VALUE, (VOID *) RE_get_value(re));
		break;
	    case TEXT_KIND:
		ISetGadgetAttributes (re->root->gl, find_gid (re),
		NULL, NULL, USE_CURRENT_VALUE, USE_CURRENT_VALUE, re->read);
		break;
	    case STRING_KIND:
		ISetGadgetAttributes (re->root->gl, find_gid (re),
		NULL, NULL, USE_CURRENT_VALUE, USE_CURRENT_VALUE, re->read);
		break;

	    case LISTVIEW_KIND:
		break;
	    }
	}
#endif
#ifdef WINDOWS
	switch (re->type)
	{
	case INTEGER_KIND:
		sprintf (re->read, "%ld", RE_get_value (re));
		SetDlgItemText (re->root->win, re->id, re->read);
	    break;

	case TEXT_KIND:
	case STRING_KIND:
	    SetDlgItemText (re->root->win, re->id, re->read);
	    break;

	case CHECKBOX_KIND:
	    SendMessage (re->g, BM_SETCHECK, RE_get_value (re), 0);
	    break;

	case CYCLE_KIND:
	    SendMessage (re->g, CB_SETCURSEL, RE_get_value (re), 0);
	    break;

	case PALETTE_KIND:
//	    break;

	case MX_KIND:
	    SendMessage (re->g, CB_SETCURSEL, RE_get_value (re), 0);
	    break;

	case SLIDER_KIND:
	    break;

	case SCROLLER_KIND:
	    break;

	case LISTVIEW_KIND:
            re->cur_pos = RE_get_value (re);
	    break;
	}
#endif
	re->update_in_progress = 0;
	memcpy (re->last_value, re->write, re->item_size);
    }
}

RE_check_in_progress (RE_resource * resource)
{
#ifdef AMIGA
    RE_resource_entry *re;

    if (!resource)
    {
	return (NULL);
    }
    re = resource->first_resource;
    while (re)
    {
	if (re->input_in_progress)
	{
#ifdef DEBUG_VERSION
	    if (re_debug)
printf ("in progress %s\n",re->name);
#endif
	    switch (re->type)
	    {
	    case INTEGER_KIND:
		RE_set_value (re, ((struct StringInfo *) re->g->SpecialInfo)->LongInt);
#ifdef DEBUG_VERSION
	    if (re_debug)
		printf ("Integer update after other gadget selected\n");
#endif
		break;
	    case STRING_KIND:
if(re->write)
		strcpy (re->write, ((struct StringInfo *) re->g->SpecialInfo)->Buffer);
#ifdef DEBUG_VERSION
	    if (re_debug)
		printf ("String update after other gadget selected\n");
#endif
		break;
	    }
	    re->input_in_progress = FALSE;
	}
	re = re->next;
    }
#endif
    return (NULL);
}


TY_int RE_process_event (RE_resource_entry * re, struct IntuiMessage * imsg)
{
    if (re && imsg)
    {
#ifdef AMIGA
	MD_GADGET_P gad = re->g;
	BOOL terminated = FALSE;
	ULONG class;
	UWORD code;

	class = imsg->Class;
	code = imsg->Code;

	if (using_v36 ())
	{
	    if ((class == GADGETDOWN) || (class == GADGETUP))
	    {
#ifdef DEBUG_VERSION
	    if (re_debug)
		printf ("checking in progress\n");
#endif
		RE_check_in_progress (re->root);
	    }

	    if (class == GADGETDOWN &&
		((re->type == STRING_KIND) ||
		 (re->type == INTEGER_KIND)))
	    {
#ifdef DEBUG_VERSION
	    if (re_debug)
		printf ("Integer/string gadget down \n");
#endif
		re->input_in_progress = TRUE;
	    }
	    if ((class == MOUSEMOVE && re->type == SCROLLER_KIND) ||
		(class == GADGETDOWN) || (class == GADGETUP) )
	    {
#ifdef DEBUG_VERSION
		if (re_debug)
		    printf ("Gadget %d %8x %d %s %d [%d]\n", re->g->GadgetID, re->g, code, re->name,
			    *(TY_int *) re->read, class);
#endif
if(class == GADGETUP)
		switch (re->type)
		{
		case BUTTON_KIND:
#ifdef DEBUG_VERSION
		    if (re_debug)
			printf ("Button '%s'.\n", re->name);
#endif
		    break;

		case INTEGER_KIND:

		    /*
		     * String/Integer gadgets report GADGETUP's, fancy or
		     * different here:
		     */
#ifdef DEBUG_VERSION
		    if (re_debug)
			printf ("Integer gadget: %ld.\n",
			 ((struct StringInfo *) gad->SpecialInfo)->LongInt);
#endif
		    RE_set_value (re, ((struct StringInfo *) gad->SpecialInfo)->LongInt);
		    re->input_in_progress = FALSE;
		    break;

		case CHECKBOX_KIND:

		    /*
		     * Checkboxes report GADGETUP's, nothing fancy or
		     * different here:
		     */
#ifdef DEBUG_VERSION
		    if (re_debug)
		    {
			printf ("%s is ", re->name);
			if (!(gad->Flags & SELECTED))
			{
			    printf ("not ");
			}
			printf ("selected.\n");
		    }
#endif
		    if (gad->Flags & SELECTED)
			RE_set_value (re, TRUE);
		    else
			RE_set_value (re, FALSE);
		    break;

		case SCROLLER_KIND:
		    /*
		     * Scroller level is in code: 
		     */
#ifdef DEBUG_VERSION
		    if (re_debug)
			printf ("Scroller Level: %d\n", code);
#endif
		    break;

		case STRING_KIND:

		    /*
		     * String gadgets report GADGETUP's, nothing fancy or
		     * different here:
		     */
#ifdef DEBUG_VERSION
		    if (re_debug)
			printf ("LVString: '%s'.\n",
			  ((struct StringInfo *) gad->SpecialInfo)->Buffer);
#endif
		    strcpy (re->write, ((struct StringInfo *) gad->SpecialInfo)->Buffer);
		    re->input_in_progress = FALSE;
		    break;

		case SLIDER_KIND:
#ifdef DEBUG_VERSION
		    if (re_debug)
			printf ("Slider %s Level: %d\n", re->name, (WORD) code);
#endif
		    RE_set_value (re, code);
break;
		case LISTVIEW_KIND:
#ifdef DEBUG_VERSION
		    if (re_debug)
			printf ("Listview %s: clicked on %d\n", re->name, code);
#endif
		    RE_set_value (re, code);
break;
		case PALETTE_KIND:
		case CYCLE_KIND:
#ifdef DEBUG_VERSION
		    if (re_debug)
			printf ("CY/CO/PA %s Level: %d\n", re->name, (WORD) code);
#endif
		    RE_set_value (re, code);
		    break;
		}
if(class == GADGETDOWN)
		switch (re->type)
		{
		case MX_KIND:
#ifdef DEBUG_VERSION
		    if (re_debug)
			printf ("MX %s Level: %d\n", re->name, (WORD) code);
#endif
		    RE_set_value (re, code);
break;
}

		if (re->process)
		    (*re->process) (re, (void*)code);

		RE_update_item (re, FALSE);
		return (terminated);
	    }
	}
	else
	{
	    if (class == ISUP_ID)
	    {
		TY_int val = (TY_int) imsg->IAddress;
#ifdef DEBUG_VERSION
		if (re_debug)
		    printf ("Gadget %d %8x %d %s %d\n",
			    re->id, re->g, val, re->name,
			    *(TY_int *) re->read);
#endif
		switch (re->type)
		{
		case BUTTON_KIND:
#ifdef DEBUG_VERSION
		    if (re_debug)
			printf ("Button '%s'.\n", re->name);
#endif

		    break;

		case INTEGER_KIND:
		    RE_set_value (re, val);
#ifdef DEBUG_VERSION
		    if (re_debug)
			printf ("Integer gadget: %ld.\n", val);
#endif
		    break;

		case CHECKBOX_KIND:
#ifdef DEBUG_VERSION
		    if (re_debug)
		    {
			printf ("%s is ", re->name);
			if (val)
			{
			    printf ("not ");
			}
			printf ("selected.\n");
		    }
#endif
		    RE_set_value (re, val);

		    break;


		case LISTVIEW_KIND:
		case SLIDER_KIND:
		case PALETTE_KIND:
		case MX_KIND:
		case CYCLE_KIND:
#ifdef DEBUG_VERSION
		    if (re_debug)
			printf ("LV/SL/CO/MX/CY: %s,  clicked on '%d'\n", re->name, val);
#endif
		    RE_set_value (re, val);
		    break;

		case SCROLLER_KIND:
		    /*
		     * Scroller level is in code: 
		     */
#ifdef DEBUG_VERSION
		    if (re_debug)
			printf ("Scroller Level: %d\n", val);
#endif
		    RE_set_value (re, val);
		    break;


		case STRING_KIND:
#ifdef DEBUG_VERSION
		    if (re_debug)
			printf ("LVString: '%s'.\n", val);
#endif
if(re->write)
		    strcpy (re->write, (char *) val);
		    break;

		}

		if (re->process)
		    (*re->process) (re, (void*)val);

		RE_update_item (re, FALSE);
		return (terminated);
	    }
	}
#endif
#ifdef WINDOWS
int evp = 0;
		TY_int re_debug = 0;
		DWORD dwIndex;

		    char ts[100];

		    if (re)
		    {
			switch (re->type)
			{
			case BUTTON_KIND:
			    if (imsg->event == BN_CLICKED)
			    {
#ifdef AAA
				if (fw)
				{
				    init_menu_test (get_root_window ());
//SetMenu (get_root_window(), fw->first_menu);
				    init_menu_test (fw->w);
				    SetMenu (fw->w, fw->first_menu);
				    op_s ("button changed menu", 0);
				}
				head = 0;
#endif
				evp = 1;
#ifdef DEBUG_VERSION
				if (re_debug)
				    op_s ("Button pushed '%s'.\n", re->name);
#endif
			    }
			    break;

/*
 * if (re_debug) 
 */
/*
 * op_s ("Integer gadget"); 
 */
/*
 * ((struct StringInfo *) gad->SpecialInfo)->LongInt); 
 */
/*
 * RE_set_value (re, ((struct StringInfo *) gad->SpecialInfo)->LongInt); 
 */
/*
 * re->input_in_progress = FALSE; 
 */
			    break;

			case CHECKBOX_KIND:
			    if (imsg->event == BN_CLICKED)
			    {
				dwIndex = SendMessage (re->g, BM_GETCHECK, 0, 0);
#ifdef DEBUG_VERSION
				if (re_debug)
				{
				    sprintf (ts, "Checkbox '%s' %s selected (%d)",
					     re->name,
					dwIndex ? "is" : "is not", dwIndex);
                                    op_s (ts, NULL);
				}
#endif
				if (dwIndex)
				    RE_set_value (re, TRUE);
				else
				    RE_set_value (re, FALSE);
				evp = 1;
			    }
			    break;

			case LISTVIEW_KIND:
			    if (imsg->event == LBN_SELCHANGE)	/*
								 * LBN_DBLCLK 
								 */
			    {
				dwIndex = SendMessage (re->g, LB_GETCURSEL, 0, 0);
				if (dwIndex != LB_ERR)
				{
				    //    SendMessage (re->g, LB_GETTEXT, (WPARAM) dwIndex, (LPARAM) ((LPSTR) ts));
#ifdef DEBUG_VERSION
				    if (re_debug)
				    {
					sprintf (ts, "Listbox %s, value %d", re->name, dwIndex);
					op_s (ts, 0);
				    }
#endif
				    RE_set_value (re, dwIndex);
            re->cur_pos = dwIndex;
				    evp = 1;
				}
			    }
			    break;

			case SLIDER_KIND:
#ifdef DEBUG_VERSION
			    if (re_debug)
				op_s ("Slider %s", re->name);
#endif
//                            SetScrollPos(fw->w,SB_CTL, wmID,S 
			    evp = 1;
			    break;

			case PALETTE_KIND:
//			    break;
// // // For now we'll treate the palette as a MX/CY

			case MX_KIND:
			case CYCLE_KIND:
			    if (imsg->event == CBN_SELCHANGE)	/*
								 * Combo Box Selection Change 
								 */
			    {
				dwIndex = SendMessage (re->g, CB_GETCURSEL, 0, 0);
#ifdef DEBUG_VERSION
				if (re_debug)
				{
				    sprintf (ts, "MX/CY %s Level: %d\n", re->name, dwIndex);
				    op_s (ts, NULL);
				}
#endif
				RE_set_value (re, dwIndex);
				evp = 1;
			    }
			    break;

			case INTEGER_KIND:
			    if (imsg->event == EN_CHANGE)
			    {
				char buf[50], *ptr, *optr, err = 0;
TY_int slen;

				*buf = 0;
				slen = SendMessage (re->g, WM_GETTEXT, sizeof (buf), buf);
#ifdef DEBUG_VERSION
				if (re_debug)
				    op_s ("LVInteger: '%s'.", re->write);
#endif
				if (slen >0 && *buf)
				{
				    optr = ptr = buf;
				    while (*ptr && slen--)
				    {
					if (isdigit (*ptr) || (*ptr == '.'))
					{
					    *optr++ = *ptr++;
					}
					else
					{
					    err = 1;
					    *ptr++;
					}
				    }
				    if (err)
					MessageBeep (-1);
else
{
TY_int new_value;
*ptr=0;
new_value = atol (buf); 
RE_set_value (re, new_value);
memcpy (re->last_value, re->write, re->item_size);
evp=1;
}
				}
break;

			    }
			case STRING_KIND:
			    if (imsg->event == EN_CHANGE)
			    {
				SendMessage (re->g, WM_GETTEXT, re->item_size, re->write);
memcpy (re->last_value, re->write, re->item_size);
evp=1;
#ifdef DEBUG_VERSION
				if (re_debug)
				    op_s ("LVString: '%s'.", re->write);
#endif
			    }
/*
 * strcpy (re->write, ((struct StringInfo *) gad->SpecialInfo)->Buffer); 
 */
			    break;
                            default:
                            return (0);
			}
return(evp);
		    }
#endif
    }
    return (0);
}


void RE_set_item (RE_resource * r, char *name, void *ptr, TY_int size)
{
    RE_resource_entry *re = RE_find_resource (name, r);
    if (!re)
	return;
#ifdef WINDOWS
if(re->type == INTEGER_KIND)
    re->write = ptr;
else
#endif
    re->read = re->write = ptr;
    re->item_size = size;

//if (re->type == LISTVIEW_KIND)
//{
//RE_set_top_lv (re, RE_get_value (re));
//RE_select_lv (re, RE_get_value (re));
//}
    FW_free (r->window, re->last_value);
    re->last_value = FW_calloc (r->window, 1, re->item_size);
    RE_update_item (re, TRUE);
}

/*
 * Create a data item and associate it with the resource.
 */
void RE_make_item (RE_resource * r, char *name, TY_int size)
{
void *p = FW_calloc(r->window, 1, size);
RE_set_item (r,name,p,size);
}

void RE_update (RE_resource * resource)
{
    RE_resource_entry *re;
    if (!resource)
    {
	error_log ("RE_update(): Given NULL resource to update");
	return;
    }
    re = resource->first_resource;
    while (re)
    {
	RE_update_item (re, FALSE);
	re = re->next;
    }
}

RE_resource_entry *RE_find_is (RE_resource * resource, TY_int id)
{
    RE_resource_entry *re;

    if (!resource)
    {
	error_log ("RE_find_is(): Given NULL resource");
	return (NULL);
    }
    re = resource->first_resource;
    while (re)
    {
	if (re->id == id)
	    return (re);
	re = re->next;
    }
    return (NULL);
}

RE_resource_entry *RE_find_shortcut (RE_resource * resource, TY_int key)
{
    RE_resource_entry *re;
    RE_resource_entry *first_lv = NULL;

    key = tolower (key);
    if (!resource)
    {
	error_log ("RE_find_shortcut(): Given NULL resource");
	return (NULL);
    }
    re = resource->first_resource;
    while (re)
    {
	if (re->key_shortcut == key || (re->key_shortcut && key == 13))
	    return (re);
	if (!first_lv && re->type == LISTVIEW_KIND)
	    first_lv = re;
	re = re->next;
    }
    if (key == RE_scroll_lv_id)
	return (first_lv);
    return (NULL);
}

void RE_set_max (RE_resource_entry * re, TY_int max)
{
#ifdef AMIGA
    if (re->type == SLIDER_KIND)
	if (using_v36 ())
	    GT_SetGadgetAttrs (re->g, re->root->win, NULL,
			       GTSL_Max, max, TAG_DONE);
#endif
}


void RE_process_shortcut (RE_resource_entry * re, struct IntuiMessage *msg, MD_WINDOW_P w)
{
#ifdef AMIGA
    MD_GADGET_P gad = re->g;

    switch (re->type)
    {
    case MX_KIND:
    case CYCLE_KIND:
	break;
    case STRING_KIND:
    case INTEGER_KIND:
	ActivateGadget (gad, w, NULL);
	break;
    case CHECKBOX_KIND:
	RE_set_value (re, !RE_get_value (re));
	break;
    default:
	{
	    msg->Class = GADGETUP;
	    RE_process_event (re, msg);
	    msg->Class = VANILLAKEY;
	    break;
	}
    }
#endif
}

void RE_disable_gadget (RE_resource_entry * re)
{
#ifdef AMIGA
    if (re)
    {
	if (using_v36 ())
	    GT_SetGadgetAttrs (re->g, re->root->win, NULL,
			       GA_Disabled, TRUE, TAG_DONE);
	else
	{
	    ISetGadgetAttributes (re->root->gl, find_gid (re),
				  GADGET_DATA_FLAG_DISABLED, GADGET_DATA_FLAG_DISABLED, USE_CURRENT_VALUE, USE_CURRENT_VALUE, re->read);
	}
    }
#endif
}

void RE_enable_gadget (RE_resource_entry * re)
{
#ifdef AMIGA
    if (re)
    {
	if (using_v36 ())
	    GT_SetGadgetAttrs (re->g, re->root->win, NULL,
			       GA_Disabled, FALSE, TAG_DONE);
	else
	{
	    ISetGadgetAttributes (re->root->gl, find_gid (re),
				  GADGET_DATA_FLAG_DISABLED, 0, USE_CURRENT_VALUE, USE_CURRENT_VALUE, re->read);
	}
    }
#endif
}

void RE_set_all_gadgets (RE_resource * resource, TY_int enabled)
{
    RE_resource_entry *re;

    re = resource->first_resource;
    while (re)
    {
if (enabled)
RE_enable_gadget (re);
else
RE_disable_gadget (re);
	re = re->next;
    }
}
TY_int RE_is_valid_gadget (RE_resource * resource, MD_GADGET_P rg)
{
    RE_resource_entry *re;
    if (!resource)
    {
	return (FALSE);
    }
    re = resource->first_resource;
    if (!re)
	return (FALSE);
    while (re)
    {
	if (re->g == rg)
	    return (TRUE);
	re = re->next;
    }
    return (FALSE);
}
