/*---------------------------------------------------------------------------
 *
 *      Title:          NICOLA Menu handling
 *
 *      File:           RE_menu.c
 *
 *      Date:           July 1993
 *
 *      Description:
 *
 *      Data sources:
 *
 *      1.0     R.J.Harrison    Implement                            1993
 *      1.1     R.J.Harrison    Add MS-Windows support          July 1993
 *      1.2     R.J.Harrison    Add Motif Support               May  1994
 *
 *      Copyright  1993 R.J.Harrison
 *---------------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#ifdef AMIGA
#include <intuition/intuition.h>
#include <proto/graphics.h>
#include <proto/intuition.h>
#include <exec/types.h>
#include <exec/lists.h>
#endif

#ifdef MOTIF
#include <Mrm/MrmAppl.h>	/*
				 * Motif Toolkit and MRM 
				 */
#endif

#include "zpl.h"

#define BIG_BLOCK
#ifdef BIG_BLOCK
static void *w_get_block (TY_int ne, TY_int esize, ZPL_WINDOW * fw, TY_int *cur_id, TY_int * free_slots, void **free_space_p, char *ty)
{
    void *free_space = *free_space_p, *new_ptr;

    if ((*free_slots <= 0) || (fw->window_id != *cur_id))
    {
#ifdef DEBUG
	if (cur_id && *free_slots)
	    printf ("%ld %s slots unused\n", *free_slots, ty);
#endif
	*cur_id = fw->window_id;
	*free_slots = ne;
	free_space = FW_calloc (fw, esize, *free_slots);
    }

    if (!free_space)
	return (0);

    new_ptr = free_space;
    *free_slots = *free_slots - 1;
    free_space = (char *) free_space + esize;
    *free_space_p = free_space;
    memset (new_ptr, 0, esize);
    return (new_ptr);
}
#else
static void *w_get_block (TY_int ne, TY_int esize, ZPL_WINDOW * fw, ZPL_WINDOW ** cur_fw, TY_int * free_slots, void **free_space_p, char *ty)
{
    return (FW_calloc (fw, ne, esize));
}
#endif

#ifdef AMIGA
static struct TextAttr TOPAZ80 =
{(STRPTR) "topaz.font", TOPAZ_EIGHTY, 0, 0};

static struct IntuiText blank_menu_text =
{2, 1, JAM2, 0, 1, NULL, "~~~~~~~~~~", NULL};

APTR
Setup_text (ZPL_WINDOW * fw, char *line)
{
    static struct IntuiText my_text =
    {0, 1, JAM2, 2, 1, NULL, "", NULL};
    static TY_int done = FALSE;
    struct IntuiText *new_text;

    static TY_int free_slots = 0;
    static TY_int cur_id=-1;
    static void *free_space;

    new_text = w_get_block (50, sizeof (*new_text), fw, &cur_id, &free_slots, &free_space, "txt");

    if (new_text)
    {
	if (!done)
	{
	    if (menu_background_not_black ())
	    {
		my_text.BackPen = brightest_colour ();
		my_text.FrontPen = darkest_colour ();
	    }
	    else
	    {
		my_text.FrontPen = brightest_colour ();
		my_text.BackPen = darkest_colour ();
	    }
	    if (my_text.FrontPen == my_text.BackPen)
		my_text.FrontPen = !my_text.BackPen;
	    done = TRUE;
	}
	*new_text = my_text;
	new_text->IText = line;
    }
    return ((APTR) new_text);
}


/*
 * The following routines adjust an entire menu system to conform to the
 * specified fonts'.
 * 
 * If a menu doesn't fit, then the return value will indicate this to the
 * caller, which should then re-adjust using a smaller font.
 */


/*
 * Adjust an individual font structure
 */
VOID
AdjustText (struct IntuiText * text, struct TextAttr * attr)
{
    struct IntuiText *nt;
    nt = text;
    while (nt)
    {
	nt->ITextFont = attr;
	nt = nt->NextText;
    }
}

/*
 * Determine the maximum width of the strip from the elements in it, and the
 * font/rp. Automatically adjust for COMMSEQ & CHECKIT.
 */
USHORT
MaxLength (struct RastPort *txtrp, struct MenuItem *fi, USHORT width)
{
    USHORT maxval = 0, textlen;
    struct MenuItem *item = fi;
    struct IntuiText *itext;
    TY_int dwidth = 0;

    while (item)
    {
	if (item->Flags & COMMSEQ)
	    dwidth = COMMWIDTH +
		TextLength (txtrp, &item->Command, 1);
	/*
	 * For Old style checkit handling, set checked here...
	 */
	item = item->NextItem;
    }
    width += dwidth;

    item = fi;
    while (item)
    {
	itext = (struct IntuiText *) item->ItemFill;
	if (itext && item->ItemFill != &blank_menu_text)
	{
	    if (item->Flags & CHECKIT)
		itext->LeftEdge = CHECKWIDTH;
	    textlen = itext->LeftEdge + TextLength (txtrp, itext->IText,
				      (LONG) strlen (itext->IText)) + width;
	    maxval = max (textlen, maxval);
	}
	item = item->NextItem;
    }
    return (maxval);
}


/*
 * Adjust the MenuItems and SubItems (subitems recurse)
 */
int
  AdjustItems (ZPL_WINDOW * fw, struct RastPort *txtrp, struct MenuItem *fi,
	       struct TextAttr *atr, USHORT wdth, USHORT hght,
	       USHORT lvl, USHORT edge, USHORT start)
{
    struct MenuItem *item = fi;
    register USHORT num = 0;
    USHORT strip_width, sub_edge;
    TY_int offset = 0, ssw;
    /*
     * Offset is to take accout of 'blank' lines. ssw = is 'sub-strip-width',
     * and is a transient information holder, to prevent multiple function
     * call access.
     */
    if (fi == NULL)
	return (0);

    strip_width = MaxLength (txtrp, item, wdth);

    while (item)
    {
	item->TopEdge = (num * hght) - lvl - offset;
	item->LeftEdge = edge;
	item->Width = strip_width;

	if (item->ItemFill)
	{
	    item->Height = hght;
	    AdjustText ((struct IntuiText *) item->ItemFill, atr);
	}
	else
	{
	    offset += hght / 2;
	}

	if (item->SubItem)
	{
	    struct IntuiText *new_itext, *last_itext = (struct IntuiText *) item->ItemFill;
	    new_itext = Setup_text (fw, "");
	    new_itext->TopEdge = last_itext->TopEdge;
	    new_itext->ITextFont = last_itext->ITextFont;
	    new_itext->FrontPen = last_itext->FrontPen;
	    new_itext->LeftEdge = ((strip_width - 1) - TextLength (txtrp, new_itext->IText, (LONG) strlen (new_itext->IText)));
	    /*
	     * wdth * 2); 
	     */


	    last_itext->NextText = new_itext;
	    item->Width = strip_width;
	    sub_edge = strip_width / 2 - wdth;
	    if (((ssw = MaxLength (txtrp, item->SubItem, wdth)) + sub_edge + start) >= get_screen_width ())
		sub_edge = -(ssw / 2);
	    AdjustItems (fw, txtrp, item->SubItem, atr, wdth, hght, 1, sub_edge, start + sub_edge);
	}
	item = item->NextItem;
	num++;
    }
    return ((num * hght) - lvl - offset);
}

BOOL
AdjustMenus (ZPL_WINDOW * fw, struct Menu * firstmenu, struct TextAttr * attr)
{
    struct RastPort textrp =
    {0};			/*
				 * Temporary RastPort 
				 */
    struct Menu *menu;
    struct TextFont *font;	/*
				 * Font to use 
				 */
    USHORT start, width, height, space;
    BOOL retval = FALSE;
    TY_int max_height = 0;

    if ((font = OpenFont (attr)))
    {
	SetFont (&textrp, font);	/*
					 * Put font into temporary RastPort 
					 */

	width = font->tf_XSize;	/*
				 * Get the Width of the Font 
				 */

	/*
	 * To prevent crowding of the Amiga key when using COMMSEQ, don't
	 * allow the items to be less than 8 pixels high.
	 */
	height = (font->tf_YSize < 8) ? 8 : font->tf_YSize;
	height++;

	start = 2;		/*
				 * Set Starting Pixel 
				 */

	/*
	 * Step thru the menu structure and adjust it 
	 */
	menu = firstmenu;
	while (menu)
	{
	    TY_int new_height;
	    menu->LeftEdge = start;
	    menu->Width = space =
		TextLength (&textrp, menu->MenuName,
			    (LONG) strlen (menu->MenuName)) + width;
	    new_height = AdjustItems (fw, &textrp, menu->FirstItem, attr, width, height, 0, 0, start);
	    max_height = max (new_height, max_height);
	    menu = menu->NextMenu;
	    start += (space + (width));
	}
	CloseFont (font);	/*
				 * Close the Font 
				 */
	/*
	 * printf ("menu sizing: %d, %d
	 * [%d,%d]\n",start,max_height,get_screen_width ()
	 * ,get_screen_height ());
	 */
	if (start >= get_screen_width () || max_height >= get_screen_height ())
	    retval = FALSE;
	else
	    retval = TRUE;
    }
    return (retval);
}
#endif

void
  enable_group (ZPL_WINDOW * fw, TY_int group)
{
    MENUITEMS *cp;

    if ((group >= 0) && (group <= get_last_group()))
	fw->active_groups[group] = TRUE;

    if (cp = fw->menulist)
    {
	while (cp)
	{
	    if (cp->group == group)
#ifdef AMIGA
		SET_BIT (cp->item->Flags, ITEMENABLED);
#endif
#ifdef WINDOWS
	    EnableMenuItem (fw->first_menu, cp->item, MF_ENABLED);
#endif
	    cp = cp->next;
	}
    }
}

void
  disable_group (ZPL_WINDOW * fw, TY_int group)
{
    MENUITEMS *cp;

    if ((group >= 0) && (group <= get_last_group()))
	fw->active_groups[group] = FALSE;

    if (cp = fw->menulist)
    {
	while (cp)
	{
	    if (cp->group == group)
#ifdef AMIGA
		STRIP_BIT (cp->item->Flags, ITEMENABLED);
#endif
#ifdef WINDOWS
	    EnableMenuItem (fw->first_menu, cp->item, MF_DISABLED | MF_GRAYED);
#endif
	    cp = cp->next;
	}
    }

}

void
  disable_all_groups (ZPL_WINDOW * fw)
{
    MENUITEMS *cp;

    if (cp = fw->menulist)
    {
	while (cp)
	{
#ifdef AMIGA
	    STRIP_BIT (cp->item->Flags, ITEMENABLED);
#endif
#ifdef WINDOWS
	    EnableMenuItem (fw->first_menu, cp->item, MF_DISABLED | MF_GRAYED);
#endif
	    cp = cp->next;
	}
    }
}

void
  finish_menu (ZPL_WINDOW * fw)
{
#ifdef AMIGA
    if (!AdjustMenus (fw, fw->first_menu, fw->current_font))
    {
	if (!AdjustMenus (fw, fw->first_menu, &TOPAZ80))
	{
	    error_log ("Cannot adjust menustrip to fit.");
	}
    }
#endif
    do_entity (fw->sub_window, WM_MENU_GROUPS, 0, 0);

    if (fw->first_menu)
#ifdef AMIGA
	SetMenuStrip (fw->w, fw->first_menu);
#endif
#ifdef WINDOWS
    SetMenu (fw->w, fw->first_menu);
#endif
#ifdef MOTIF
/*
 * Now manage all of the widgets...
 */
    ;
#endif

}

void
  begin_menu ()
{
}


void
  steal_menu (ZPL_WINDOW * onto, ZPL_WINDOW * from)
{
#ifdef AMIGA
    SetMenuStrip (onto->w, from->first_menu);
    onto->first_menu = from->first_menu;
    onto->menulist = from->menulist;
    onto->menu_from = from;
#endif
#ifdef WINDOWS
    SetMenu (onto->w, from->first_menu);
    onto->first_menu = from->first_menu;
    onto->menulist = from->menulist;
    onto->menu_from = from;
#endif
}

void
  add_item_to_window (ZPL_WINDOW * fw, MD_MENU_ITEM item,
		      TY_int group, void (*func) ())
{
    MENUITEMS *cp, *nmi;

    static TY_int free_slots = 0;
    static TY_int cur_id = 0;
    static void *free_space;

    nmi = w_get_block (50, sizeof (MENUITEMS), fw, &cur_id, &free_slots, &free_space, "itw");

    if (cp = fw->menulist)
    {
#ifdef AAA
	while (cp->next)
	    cp = cp->next;
	cp->next = nmi;
#else
	fw->menulist = nmi;
	nmi->next = cp;
#endif
    }
    else
	fw->menulist = nmi;

    nmi->item = item;
    nmi->group = group;
    nmi->func = func;

}

void *
  find_menuitem (ZPL_WINDOW * fw, MD_MENU_ITEM mi)
{
    MENUITEMS *cp;

    if (cp = fw->menulist)
    {
	while (cp)
	{
	    if (cp->item == mi)
		return (cp);
	    cp = cp->next;
	}
    }
    return (0);
}


void *find_item_routine (ZPL_WINDOW * fw, MD_MENU_ITEM mi)
{
    MENUITEMS *cp = find_menuitem (fw, mi);

    if (cp)
	return (cp->func);
    else
	return (NULL);

}

void
  do_nothing (TY_int code, ZPL_WINDOW * fw)
{
    return;
}

static void call_menu_func (TY_int code, ZPL_WINDOW * fw, void (*function) (TY_int, void *, MD_MENU_ITEM), MD_MENU_ITEM mi)
{
    if (function)
    {
	FW_remove_cursor (fw);
        set_close_allow (fw, FALSE);
	FW_output_all_text (fw);

	if (fw->menu_from)
	    (*function) (code, fw->menu_from, mi);
	else
	    (*function) (code, fw, mi);
	/*
	 * There are some functions during which this window can be closed
	 * so protect ourselves from this...
	 */
	if (find_window (fw->w) && !BIT_SET (fw->type, FW_TYPE_GADTOOLS))
	{
	    FW_reset_scroll (fw);
	    FW_cursor_move (fw);
	    FW_clear_more (fw);
	}
        set_close_allow (fw, TRUE);
    }
}

#ifdef MOTIF
/*
 * Adds a pulldown menu onto a widget (menubar/menuitem)
 */
Widget
W_add_menu (Widget parent, char *title, char sc)
{
    Widget pulldown, button;
    Arg args[4], narg;

    pulldown = XmCreatePulldownMenu (parent, "p1", 0, 0);
    XtSetArg (args[0], XmNsubMenuId, pulldown);
    XtSetArg (args[1], XmNlabelString, XmStringCreate (title, XmSTRING_DEFAULT_CHARSET));
    XtSetArg (args[2], XmNmnemonic, sc);
    button = XmCreateCascadeButton (parent, title, args, 3);
    XtManageChild (button);
    return pulldown;
}

static
  menu_call_func (Widget w, MD_CALLBACK_ITEM * cbs, void *reason)
{
    call_menu_func (0, cbs->fw, cbs->func, w);
}


/*
 * Creates a menu item button
 */
Widget
W_add_item (void *fw, Widget menu, char *text, char sc, void (*func) (TY_int, ZPL_WINDOW *))
{
    Widget button;
    Arg args[4], narg;

    XtSetArg (args[0], XmNlabelString, XmStringCreate (text, XmSTRING_DEFAULT_CHARSET));
    XtSetArg (args[1], XmNmnemonic, sc);
    button = XmCreatePushButtonGadget (menu, text, args, 2);
    if (func)
    {
	MD_CALLBACK_ITEM *cbs = FW_calloc (fw, sizeof (*cbs), 1);
	cbs->func = func;
	cbs->fw = fw;
	XtAddCallback (button, XmNactivateCallback, menu_call_func, cbs);
    }
    XtManageChild (button);
    return button;
}
#endif

MD_MENU *
  add_menu (ZPL_WINDOW * fw, MD_MENU * cur_menu, USHORT flags, char *title)
{
#ifdef AMIGA
    MD_MENU *new_menu;

    static TY_int free_slots = 0;
    static TY_int cur_id = 0;
    static void *free_space;

    new_menu = w_get_block (6, sizeof (*new_menu), fw, &cur_id, &free_slots, &free_space, "mnu");

    if (!new_menu)
	return cur_menu;

    if (cur_menu)
	cur_menu->NextMenu = new_menu;
    else
	fw->first_menu = new_menu;

    new_menu->Flags = flags;
    new_menu->MenuName = title;	/*
				 * text for menu-bar display 
				 */

#endif
#ifdef WINDOWS
    MD_MENU *new_menu;
    if (!cur_menu)
    {
	cur_menu = new_menu = FW_calloc (fw, sizeof (*cur_menu), 1);
	cur_menu->menu = CreateMenu ();
	fw->first_menu = cur_menu->menu;
    }
    cur_menu->title = title;
    new_menu = cur_menu;
#endif
#ifdef MOTIF
    MD_MENU *new_menu;
    if (!cur_menu)
    {
	Widget hw;
	Arg args[4], narg;
	cur_menu = new_menu = FW_calloc (fw, sizeof (*cur_menu), 1);

	XtSetArg (args[0], XmNtopAttachment, XmATTACH_FORM);
	XtSetArg (args[1], XmNrightAttachment, XmATTACH_FORM);
	XtSetArg (args[2], XmNleftAttachment, XmATTACH_FORM);
	cur_menu->menu_bar = XmCreateMenuBar (fw->top_widget, "menubar", args, 3);

	hw = W_add_item (fw, cur_menu->menu_bar, "Help", 'H', 0);

	XtSetArg (args[0], XmNmenuHelpWidget, (XtArgVal) hw);
	XtSetValues (cur_menu->menu_bar, args, 1);

	fw->first_menu = cur_menu;
    }
    cur_menu->pulldown = W_add_menu (cur_menu->menu_bar, title, 0);
#endif
    return (new_menu);
}

void
  init_item (MD_MENU * cur_menu, MD_MENU_ITEM cur_item)
{
#ifdef AMIGA
    cur_menu->FirstItem = cur_item;	/*
					 * pointer to first item in list 
					 */
#endif
#ifdef WINDOWS
    /*
     * Add the pop-up menu to the main menu. 
     */
    AppendMenu (cur_menu->menu, MF_ENABLED | MF_POPUP, (UINT) cur_item, cur_menu->title);

#endif
#ifdef MOTIF
/*
 * Setup the Cur_Item to point to the menu root. This is
 * needed for each menu, to add the buttons onto the pulldown
 * widget
 */
    cur_item->menu_root = cur_menu;
#endif
}


MD_MENU_ITEM
add_item (ZPL_WINDOW * fw, MD_MENU_ITEM cur_item,
	  TY_int group, char *src_text,
	  TY_int select, char command, MD_MENU_ITEM subitem,
	  void (*func) (TY_int, ZPL_WINDOW *))
{
    if (func == do_nothing && !subitem && src_text)
	return (cur_item);
    else
    {
#ifdef AMIGA
	USHORT flags = ITEMTEXT | ITEMENABLED | HIGHCOMP;
	APTR text;
	MD_MENU_ITEM new_item;

	static TY_int free_slots = 0;
	static TY_int cur_id = 0;
	static void *free_space;

	new_item = w_get_block (50, sizeof (*new_item), fw, &cur_id, &free_slots, &free_space, "itm");
	if (!new_item)
	    return 0;
	if (src_text)
	    text = Setup_text (fw, src_text);
	else
	    text = NULL;

	if (cur_item)
	{
	    cur_item->NextItem = new_item;
	}

	if (command)
	    flags |= COMMSEQ;

	new_item->Flags = flags;
	new_item->ItemFill = text;
	new_item->SelectFill = select;
	new_item->Command = command;
	new_item->SubItem = subitem;
	if (func && func != do_nothing)
	    add_item_to_window (fw, new_item, group, func);
	return (new_item);
#endif
#ifdef WINDOWS
	MD_MENU_ITEM new_item;
	static TY_int unique_id = MD_FIRST_MENU;
	if (!cur_item)
	    cur_item = CreatePopupMenu ();
	if (src_text)
	{
	    unique_id++;
	    if (subitem)
		AppendMenu (cur_item, MF_POPUP, subitem, src_text);
	    else
	    {
		AppendMenu (cur_item, MF_ENABLED | MF_STRING, unique_id, src_text);
		if (func && func != do_nothing)
		    add_item_to_window (fw, unique_id, group, func);
	    }
	}
	else
	    AppendMenu (cur_item, MF_SEPARATOR, -1, NULL);
	new_item = cur_item;
	return (new_item);
#endif
#ifdef MOTIF
	MD_MENU_ITEM new_item = FW_calloc (fw, sizeof (*new_item), 1);

	memset (new_item, 0, sizeof (*new_item));

	if (cur_item)
	{
	    new_item->menu_root = cur_item->menu_root;
	}
	new_item->mw = W_add_item (new_item->menu_root->pulldown, src_text, command);

	add_item_to_window (fw, new_item, group, func);

	return (new_item);

#endif
    }
}

USHORT
process_menu (USHORT code, ZPL_WINDOW * fw)
{
    MD_MENU_ITEM mi;
    void (*function) (TY_int, void *, MD_MENU_ITEM);
#ifdef AMIGA
    mi = ItemAddress (fw->first_menu, (long) code);
    function = find_item_routine (fw, mi);
#endif
#ifdef WINDOWS
    if (code < MD_FIRST_MENU)
	return (0);
    function = find_item_routine (fw, code);
#endif
    if (function)
	call_menu_func (code, fw, function, mi);
    else
	return 0;
#ifdef AMIGA
    return (mi->NextSelect);
#endif
#ifdef WINDOWS
    return (1);
#endif
}

