/*---------------------------------------------------------------------------
 *
 *      Title:          NICOLA scrollback handling
 *
 *      File:           sb.c
 *
 *      Date:           1993
 *
 *      Description:
 *          This module handles the scrollback part of the window.  The scrollback is
 *          also used for the buffered more?  prompt output.
 *
 *      Rev: 1.0
 *      1.0     R.J.Harrison    Implement                           March 1993
 *
 *      Copyright  1993 R.J.Harrison
 *---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "zpl.h"

#define FW_GET_ID(fw) ((TY_int)fw)
/*
 * #define SB_DEBUG
 */

void FW_init_sb (ZPL_WINDOW * fw)
{
    QDB_GRAN_ITEM *qit = &fw->scrollback.list;
    memset (qit, 0, sizeof (*qit));
    init_gi_item (qit, sizeof (SCROLLBACK_ITEM), PR_get_granularity ());
}

void FW_free_sb (ZPL_WINDOW * fw)
{
    QDB_GRAN_ITEM *qit = &fw->scrollback.list;
    if (qit->init)
{
    SCROLLBACK_ITEM *list = qit->ptr;
if(list)
{
int i;
    for (i = 0; i < fw->scrollback.list.size; i++)
    {
	if (list[i].line)
	{
	    free (list[i].line);
	    list[i].line = 0;
	}
}
}
	free_gi_item (qit);
}
}

void FW_set_sb (ZPL_WINDOW * fw)
{
#ifdef AMIGA
    RE_set_scrollbar (fw->scrollback.scrollbar, fw->scrollback.top, fw->scrollback.list.size, fw->t_height);
#endif
#ifdef WINDOWS
    SetScrollRange (fw->w, SB_VERT, 0, fw->scrollback.list.size, FALSE);
    SetScrollPos (fw->w, SB_VERT, fw->scrollback.top, TRUE);
#endif
}

/*
 * position the viewing 'window' to have the specified top.
 * - firstly discover how far away the required top is from the current top
 *  -- if it is less than a third of the window then scroll.
 */
void FW_draw_sb (ZPL_WINDOW * fw, TY_int top)
{
    QDB_GRAN_ITEM *qit = &fw->scrollback.list;
    SCROLLBACK_ITEM *list;
    TY_int lines_to_move, ptr, ltm, offset;

    if ((top != fw->scrollback.top))
    {
	lines_to_move = top - fw->scrollback.top;
	list = qit->ptr;

	FW_remove_cursor (fw);
#ifdef SB_DEBUG
	printf ("move window to %d from %d (%d)\n", top, fw->scrollback.top, lines_to_move);
#endif

#ifndef WIN_BUF
	if (abs (lines_to_move) < fw->t_height / 3)
	{
	    if (lines_to_move < 0)
	    {
		FW_scroll_up (fw, abs (lines_to_move));
		FW_move (fw, 0, 0);
		offset = 0;	/*
				 * Offset from top 
				 */
	    }
	    else
	    {
		FW_scroll (fw, abs (lines_to_move));
		offset = fw->t_height - abs (lines_to_move);
		FW_move (fw, 0, offset);
		/*
		 * printf ("offset %d\n", offset);
		 */
	    }
	    ltm = abs (lines_to_move);
	    for (ptr = 0; ptr < ltm; ptr++)
	    {
		TY_int ind = top + ptr + offset;
		FW_erase_to_eol (fw);
		if (ind < qit->size)
		{
		    FW_colour (fw, list[ind].colour);
		    FW_erase_to_eol (fw);
		    if (list[ind].line)
   		        FW_puts_hl (fw, list[ind].line);
		    list[ind].output = TRUE;
		}
		else
		    FW_puts (fw, "\n");

	    }
	}
	else
#endif
	{
#ifdef SB_DEBUG
	    printf ("Redrawing window\n");
#endif
	    /*
	     * FW_cls (fw);
	     */
	    FW_move (fw, 0, 0);
	    ltm = min (qit->size, top + fw->t_height);
	    for (ptr = top; ptr < ltm; ptr++)
	    {
		list[ptr].output = TRUE;
		FW_erase_to_eol (fw);
		FW_colour (fw, list[ptr].colour);
	    if (list[ptr].line)
		FW_puts_hl (fw, list[ptr].line);
	    }
	}
	fw->scrollback.top = top;
	FW_set_sb (fw);
	FW_move (fw, 0, fw->t_height);
	FW_erase_to_eol (fw);
    }
}

void FW_redraw_sb (ZPL_WINDOW * fw)
{
    if (fw && BIT_SET (fw->type, FW_SCROLLBACK))
    {
	TY_int top = fw->scrollback.top;
	fw->scrollback.top = -100;
	FW_cls (fw);
	FW_draw_sb (fw, top);
    }
}

void FW_draw_new_sb (ZPL_WINDOW * fw)
{
    QDB_GRAN_ITEM *qit = &fw->scrollback.list;
    SCROLLBACK_ITEM *list;
    TY_int ptr, ltm;

    if (fw->scrollback.new_output)
    {
	TY_int line = 0;
	list = qit->ptr;
	FW_remove_cursor (fw);
	ltm = min (qit->size, fw->scrollback.top + fw->t_height);
#ifdef SB_DEBUG
	printf ("draw_new_sb: top=%d, ltm=%d\n", fw->scrollback.top, ltm);
#endif
	for (ptr = fw->scrollback.top; ptr < ltm; ptr++)
	{
	    if (!list[ptr].output)
	    {
		FW_move (fw, 0, line);
		FW_erase_to_eol (fw);
		FW_colour (fw, list[ptr].colour);
	    if (list[ptr].line)
		FW_puts_hl (fw, list[ptr].line);
		list[ptr].output = TRUE;
	    }
	    line++;
	}
    }
}


/*
 * This routine adds a line onto the scrollback.
 * - If a more? prompt is currently active then the list is allowed to grow
 *   past the size in the prefs...
 * - Otherwise lines will be removed from the beginning to make space for
 *   the new lines.
 */
void FW_add_sb (ZPL_WINDOW * fw, char *line)
{
    if (line && *line)
    {
	GENERAL_PREFS *gp = PR_get_general_prefs ();
	QDB_GRAN_ITEM *qit = &fw->scrollback.list;
	SCROLLBACK_ITEM *list;
	TY_int doing_more;

	if (!qit->init)
	    init_gi_item (qit, sizeof (SCROLLBACK_ITEM), gp->scrollback_size / 4);

	if (((fw->scrollback.displayed_upto + fw->t_height) < fw->scrollback.display_position) ||
	    BIT_SET (fw->type, FW_WAITING_FOR_MORE))
	    doing_more = TRUE;
	else
	    doing_more = FALSE;

	list = qit->ptr;

	if (!doing_more && (qit->size > gp->scrollback_size))
	{
	    int i, lines_to_remove = qit->size - gp->scrollback_size;

	    if (lines_to_remove < gp->scrollback_remove)
		lines_to_remove = gp->scrollback_remove;

#ifndef STATIC_SB
	    for (i = 0; i < lines_to_remove; i++)
	    {
		if (list[i].line)
		{
		    free (list[i].line);
		    list[i].line = 0;
		}
	    }
#endif
	    QDB_delete_gi (qit, 0, lines_to_remove);
	    fw->scrollback.displayed_upto -= lines_to_remove;
	    fw->scrollback.top -= lines_to_remove;
/*
 * fw->scrollback.more_upto -= lines_to_remove;
 */
	    FW_set_sb (fw);
	}

	if (!strchr (line, '\n'))
	    strcat (line, "\n");

	check_gi_alloc (qit);
	list = qit->ptr;
	qit->changed = TRUE;

	list[qit->size].output = FALSE;
	list[qit->size].colour = fw->colour;
#ifdef STATIC_SB
	strcpy (list[qit->size++].line, line);
	if (qit->size < qit->max_size)
	    strcpy (list[qit->size].line, "zz");
#else
#ifdef WIERD_SHIT
	if (list[qit->size].line && (strlen (line) >= list[qit->size].size))
	{
	    free (list[qit->size].line);
	    list[qit->size].size = 0;
	    list[qit->size].line = 0;
	}
	if (!list[qit->size].line)
	{
	    list[qit->size].line = savestr (line);
	    list[qit->size].size = strlen (line);
	    qit->size++;
	}
	else
	    strcpy (list[qit->size++].line, line);
#else
	list[qit->size++].line = savestr (do_line_highlighting(fw->colour, line));
#endif /*
        * wierd
        */
#endif

	fw->scrollback.new_output = TRUE;
	fw->scrollback.display_position = max (0, qit->size - fw->t_height);
    }
}

/*
 * Output the buffered text, taking into account the more? prompt (if active).
 * If the more? prompt isn't active, then all this has to do is 'FW_draw_sb',
 * otherwise FW_draw_sb is used to display the first line of text,
 * The 'batch scroll flag' is used to determine how many lines may be output
 * in one go.
 * more action is - 1 for 'ok to more'  0 for 'not ok' and '-1' for output to end/
 */
static TY_int output_sb (ZPL_WINDOW * fw, TY_int more_action)
{
    GENERAL_PREFS *gp = PR_get_general_prefs ();
    TY_int max_lines_to_output, do_more;
    TY_int output_finished = TRUE;

    if (!more_action && BIT_SET (fw->type, FW_WAITING_FOR_MORE))
	return (TRUE);

    if (more_action || fw->scrollback.new_output)
    {

#ifdef SB_DEBUG
	printf ("\noutput_sb:  %d (mu=%d), %d %d\n",
		fw->scrollback.displayed_upto, fw->scrollback.more_upto,
		fw->scrollback.display_position, more_action);

#endif

	if (gp->batch_scroll)
	{
	    if (gp->batch_scroll == 2)
		max_lines_to_output = fw->t_height;
	    else
		max_lines_to_output = fw->t_height / 3;
	}
	else
	    max_lines_to_output = 1;

	if (more_action == -1 ||
	  (gp->batch_scroll && (!gp->use_more_prompt || fw->more_disabled)))
	{
	    FW_draw_sb (fw, fw->scrollback.display_position);
	    FW_draw_new_sb (fw);
	    fw->scrollback.displayed_upto = fw->scrollback.display_position;
	}
	else
	{
	    TY_int i, end_display;
	    do_more = FALSE;

	    if (fw->more_disabled || !gp->use_more_prompt)
	    {
		end_display = fw->scrollback.display_position;
	    }
	    else
	    {
		end_display = fw->scrollback.displayed_upto + (fw->t_height - 1);
	    }
/*
 * Fudge for the 'top of scroll' with more 
 */
	    if (!fw->scrollback.displayed_upto &&
		(fw->scrollback.display_position > end_display))
	    {
		end_display = 1;
		do_more = 1;
	    }

	    if (end_display > fw->scrollback.display_position)
		end_display = fw->scrollback.display_position;

#ifdef SB_DEBUG
	    printf ("display_proper :  dp=%d du=%d mu=%d ed=%d more=%d\n",
	     fw->scrollback.display_position, fw->scrollback.displayed_upto,
		    fw->scrollback.more_upto, end_display, do_more);

#endif
	    for (i = fw->scrollback.displayed_upto; i < end_display; i += max_lines_to_output)
	    {
#ifdef SB_DEBUG
		printf ("LO: %d\n", i);
#endif
		FW_draw_sb (fw, i);
		FW_draw_new_sb (fw);
	    }
#ifdef SB_DEBUG
	    printf ("After loop, doing %d\n", end_display);
#endif

	    FW_draw_sb (fw, end_display);
	    FW_draw_new_sb (fw);

	    if (end_display < fw->scrollback.display_position ||
		fw->lines_printed > fw->t_height)
		do_more = TRUE;

	    if (do_more && !fw->more_disabled)
	    {
		FW_do_more (fw, TRUE);
		output_finished = FALSE;
	    }
	    fw->scrollback.displayed_upto = end_display;
	}
	if (!more_action)
	    fw->scrollback.new_output = FALSE;
    }
    return (output_finished);
}

TY_int FW_output_sb (ZPL_WINDOW * fw, TY_int ok_to_more)
{
    return (output_sb (fw, ok_to_more));
}

TY_int FW_output_all_sb (ZPL_WINDOW * fw)
{
    return (output_sb (fw, -1));
}

void FW_view_sb (ZPL_WINDOW * fw)
{
    TY_int i;
    QDB_GRAN_ITEM *qit = &fw->scrollback.list;
    SCROLLBACK_ITEM *list;

    list = qit->ptr;
    for (i = 0; i < qit->size; i++)
	printf ("%3d [%2d]: %s", i, strlen (list[i].line), list[i].line);
}

void FW_remove_sb_after_end (ZPL_WINDOW * fw)
{
#ifndef STATIC_SB
    SCROLLBACK_ITEM *list;
    int i;
    list = fw->scrollback.list.ptr;
    for (i = fw->scrollback.displayed_upto + fw->t_height; i < fw->scrollback.list.size; i++)
    {
	if (list[i].line)
	{
	    free (list[i].line);
	    list[i].line = 0;
	}
    }
#endif
    fw->scrollback.list.size = min (fw->scrollback.list.size, fw->scrollback.displayed_upto + fw->t_height);
    fw->scrollback.display_position = fw->scrollback.displayed_upto;
    FW_set_sb (fw);
}

static void process_scrollbar (RE_resource_entry * re, UWORD id)
{
    if (re->root->window)
    {
	FW_draw_sb (re->root->window, id);
    }
}

static TY_int setup_scrollbar (ZPL_WINDOW * fw)
{
#ifdef AMIGA
    RE_resource_entry *re_sb;
    if (re_sb = RE_find_resource ("scrollbar", fw->resources))
    {
	RE_set_process (re_sb, process_scrollbar, (void *) fw);
	fw->scrollback.scrollbar = re_sb;
	return (TRUE);
    }
    return (FALSE);
#else
    return (TRUE);
#endif
}

/*
 * Move the scrollbar relative to its current position.
 */
void FW_move_sb (ZPL_WINDOW * fw, TY_int num_lines)
{
    TY_int new_position = fw->scrollback.top + num_lines;
    if (new_position < 0)
	new_position = 0;

    if (new_position > (fw->scrollback.list.size - fw->t_height + 1))
	new_position = (fw->scrollback.list.size - fw->t_height + 1);

    FW_draw_sb (fw, new_position);

}

void FW_add_scrollbar (ZPL_WINDOW * fw)
{
    SET_BIT (fw->type, FW_TYPE_GADTOOLS | FW_SCROLLBACK);
#ifdef AMIGA
    fw->resources = RE_add_scrollbar (fw->resources, fw->w, fw);
    RE_add_resources (fw->resources, fw->w);
#endif
#ifdef WINDOWS
SendMessage(fw->w, WM_SETFONT, (WPARAM) FW_get_fixed_font(), 0);
#endif
    if (setup_scrollbar (fw))
	FW_init_sb (fw);
    FW_reset_scroll (fw);
}

void FW_change_sb_size (ZPL_WINDOW * fw)
{
    if (!fw)
	return;
#ifdef AMIGA
    if (!using_v36 ())
	FW_redraw_frame (fw);

    fw->resources = (void *) RE_resize_scrollbar (fw->resources, fw);
    setup_scrollbar (fw);
    FW_redraw_sb (fw);
    FW_set_sb (fw);

    if (using_v36 ())
	FW_redraw_frame (fw);
#else
    setup_scrollbar (fw);
    FW_redraw_sb (fw);
    FW_set_sb (fw);
#endif
}

void FW_clear_sb (ZPL_WINDOW * fw)
{
    FW_cls (fw);
    fw->scrollback.list.size = 0;
    FW_set_sb (fw);
    setup_scrollbar (fw);
}
