#include <exec/types.h>
#include <exec/io.h>
#include <exec/memory.h>
#include <devices/timer.h>

#include <clib/exec_protos.h>
#include <clib/alib_protos.h>
#include <clib/dos_protos.h>
#include <clib/intuition_protos.h>

#include <time.h>

static void set_time (ULONG amiga_secs)
{
    struct timerequest *TimerIO;
    struct MsgPort *TimerMP;

    if (TimerMP = CreatePort (0, 0))
    {
	if (TimerIO = (struct timerequest *)
	    CreateExtIO (TimerMP, sizeof (struct timerequest)))
	{
	    /*
	     * Open with UNIT_VBLANK, but any unit can be used 
	     */
	    if (!OpenDevice (TIMERNAME, UNIT_VBLANK, (struct IORequest *) TimerIO, 0L))
	    {
		/*
		 * Set the New time...
		 */
		TimerIO->tr_time.tv_micro = 0;
		TimerIO->tr_time.tv_secs = amiga_secs;

		/*
		 * Issue the command and wait for it to finish.
		 */
		TimerIO->tr_node.io_Command = TR_SETSYSTIME;
		DoIO ((struct IORequest *) TimerIO);

		/*
		 * Now close the timer device 
		 */
		CloseDevice ((struct IORequest *) TimerIO);
	    }
	    /*
	     * Delete the I/O request structure 
	     */
	    DeleteExtIO (TimerIO);
	}
	/*
	 * Delete the port 
	 */
	DeletePort (TimerMP);
    }
}

void set_system_time (time_t new_time)
{
    /*
     * Subtract 8 years worth of seconds, including 2 leap years 
     */
    new_time -= ((6 * 365) + (2 * 366)) * (60 * 60 * 24)-3600;
    set_time (new_time);
}
